"""
Reactive Reality Machine Learning Config System - ConfigHooksMixin object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import os
from typing import Any, Callable, List, Optional, Tuple, Union

from ..yaecs_utils import (ConfigDeclarator, Hooks, Priority, VariationDeclarator,
                           assign_order, assign_yaml_tag, hook)

YAECS_LOGGER = logging.getLogger(__name__)


class ConfigHooksMixin:
    """ Hooks Mixin class for YAECS configurations. Implements processing functions whose name start with `register_as_`
    and are decorated by the yaecs_utils.hook decorator. Users can use those processing either as pre- or
    post-processing functions, which will have the added effect of tagging the processed parameters as playing a certain
    pre-defined role in the config """

    __getattribute__: Callable[[str], Any]
    get_processed_param_name: Callable[[bool], str]
    get_variation_name: Callable[[], str]
    init_from_config: Callable[[ConfigDeclarator], None]
    _configuration_variations: List[Tuple[str, List[ConfigDeclarator]]]
    _configuration_variations_names: List[Tuple[str, List[str]]]
    _grids: List[List[str]]
    _nesting_hierarchy: List[str]

    def __init__(self, *args, **kwargs):
        self._hooks = {}
        super().__init__(*args, **kwargs)

    def add_currently_processed_param_as_hook(self, hook_name: str) -> None:
        """
        Used within _ConfigurationBase._process_parameter to add the param currently being processed as a hook with
        given name. Instead of using this directly, users should consider decorating their hooking function with the
        `yaecs_utils.hook` decorator.

        :param hook_name: name of the hook to add.
        """
        parameter = self.get_processed_param_name(full_path=True)
        if hook_name not in self.get_hook():
            self._hooks[hook_name] = []
        if parameter not in self._hooks[hook_name]:
            self._hooks[hook_name].append(parameter)

    def get_experiment_path(self) -> str:
        """
        Returns the value of the parameter registered as the experiment path.

        :raises RuntimeError: when no experiment path has been registered in the config
        :raises RuntimeError: when more than one experiment path has been registered in the config
        :return: experiment path
        """
        path = self.get_hook("experiment_path")
        if not path:
            raise RuntimeError("No experiment path was registered. Please use self.register_as_experiment_path as a "
                               "post-processing on a parameter.")
        if len(path) > 1:
            raise RuntimeError("The self.register_as_experiment_path post-processing was used on more than one "
                               f"parameter : {path}.")
        return self[path[0]]

    def get_hook(self, hook_name: Optional[str] = None) -> Hooks:
        """
        Returns the parameters registered with given hook.

        :param hook_name: name of the hook, or None to get the dict of all hooked parameters
        :return: list of hooked parameter names
        """
        if hook_name is None:
            return self._hooks
        return self._hooks[hook_name] if hook_name in self._hooks else []

    @hook("additional_config_file")
    @assign_order(Priority.OFTEN_LAST)  # processing this param after this function makes it unclear which file was used
    @assign_yaml_tag("additional_config_file", "pre", "Optional[Union[str,List[str]]]")
    def register_as_additional_config_file(self, path: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Pre-processing function used to register the corresponding parameter as a path to another config file. The new
        config file will then also be used to build the config currently being built.
        Priority : OFTEN_LAST (10)
        YAML tag : additional_config_file

        :param path: config's path or list of paths
        :return: the same path as the input once the parameters from the new config have been added
        """
        if isinstance(path, list):
            for individual_path in path:
                self.init_from_config(individual_path)
        else:
            self.init_from_config(path)
        return path

    @hook("config_variations")
    @assign_order(Priority.INDIFFERENT)  # does not depend on or change parameter value, only processes parameter name
    @assign_yaml_tag("config_variations", "pre", "Optional[Union[List[Union[str,dict]],Dict[Union[str,dict]]]]")
    def register_as_config_variations(
            self, variation_to_register: Optional[VariationDeclarator]) -> Optional[VariationDeclarator]:
        """
        Pre-processing function used to register the corresponding parameter as a variation for the current config.
        Please note that config variations need to be declared in the root config.
        Priority : INDIFFERENT (0)
        YAML tag : config_variations

        :param variation_to_register: list of configs
        :return: the same list of configs once the configs have been added to the internal variation tracker
        :raises RuntimeError: register_as_config_variations is called outside _pre_process_parameter
        :raises RuntimeError: variation name invalid in sub-config
        :raises TypeError: type of variation is not list or dict of configs
        """
        name = self.get_processed_param_name(full_path=False)

        def _is_single_var(single):
            return isinstance(single, (str, dict))

        def _add_to_variations(variations, names=None):
            if variations:
                for index, variation in enumerate(self._configuration_variations):
                    if variation[0] == name:
                        self._configuration_variations.pop(index)
                        break
                self._configuration_variations.append((name, variations))
                if names is None:
                    self._configuration_variations_names.append((name, [str(i) for i in list(range(len(variations)))],))
                else:
                    self._configuration_variations_names.append((name, names))

        if self._nesting_hierarchy:
            raise RuntimeError(f"Variations declared in sub-configs are invalid ({name}).\n"
                               "Please declare all your variations in the main config.")
        if (isinstance(variation_to_register, dict)
                and all(_is_single_var(potential_single) for potential_single in variation_to_register.values())):
            _add_to_variations(list(variation_to_register.values()), names=list(variation_to_register.keys()),)
        elif (isinstance(variation_to_register, list)
              and all(_is_single_var(potential_single) for potential_single in variation_to_register)):
            _add_to_variations(variation_to_register)
        elif variation_to_register is not None:
            raise TypeError("Variations parsing failed : variations parameters must "
                            "be a list of configs or a dict containing only configs. "
                            f"Instead, got : {variation_to_register}")

        return variation_to_register

    @hook("experiment_path")
    @assign_order(Priority.OFTEN_FIRST)  # other params might depend on this being done (eg. for folder_in_experiment)
    @assign_yaml_tag("experiment_path", "post", "Optional[str]")
    def register_as_experiment_path(self, path: Optional[str]) -> Optional[str]:
        """
        Pre-processing function used to register the corresponding parameter as the folder used for the current
        experiment. This will automatically create the relevant folder structure and append an experiment index at the
        end of the folder name to avoid any overwriting. The path needs to be either None or an empty string (in which
        case the pre-processing does not happen), or an absolute path, or a path relative to the current working
        directory.
        Priority : OFTEN_FIRST (-10)
        YAML tag : experiment_path

        :param path: None, '', absolute path or path relative to the current working directory
        :return: the actual created path with its appended index
        """
        if not path:
            return path
        folder, experiment = (os.path.dirname(path), os.path.basename(path))
        if not folder:
            folder = "."
        os.makedirs(folder, exist_ok=True)
        experiments = [i for i in os.listdir(folder) if i.startswith(experiment)]
        run_ids = [-1]
        for exp_name in experiments:
            try:
                run_ids.append(int(exp_name.split("_")[-1]))
            except ValueError:
                pass
        new_folder = not os.getenv('PICKUP') and not os.getenv('NODE_RANK')
        if self.get_variation_name() is None:
            path = os.path.join(folder, f"{experiment}_{max(run_ids) + int(new_folder)}")
        else:
            path = os.path.join(folder, f"{experiment}_{max(run_ids)}", self.get_variation_name())
        if new_folder:
            os.makedirs(path, exist_ok=True)
        return path

    @hook("grid")
    @assign_order(Priority.INDIFFERENT)  # does not depend on or change parameter value, only processes parameter name
    @assign_yaml_tag("grid", "pre", "Optional[List[str]]")
    def register_as_grid(self, list_to_register: Optional[List[str]]) -> Optional[List[str]]:
        """
        Pre-processing function used to register the corresponding parameter as a grid for the current config. Grids are
        made of several parameters registered as variations. Instead of adding the variations in those parameters to the
        list of variations for this config, a grid will be created and all its components will be added instead.
        Priority : INDIFFERENT (0)
        YAML tag : grid

        :param list_to_register: list of parameters composing the grid
        :raises TypeError: list_to_register is not recognised as a valid grid
        :return: the same list of parameters once the grid has been added to the internal grid tracker
        """
        if isinstance(list_to_register, list) and all(isinstance(param, str) for param in list_to_register):
            self._grids.append(list_to_register)
        elif list_to_register is not None:
            raise TypeError("Grid parsing failed : unrecognised grid declaration : "
                            f"{list_to_register}")
        return list_to_register

    @hook("mode")
    @assign_order(Priority.INDIFFERENT)  # does not depend on or change parameter value, only processes parameter name
    @assign_yaml_tag("mode", "pre", "(str,[str])")
    def register_as_mode(self, mode: str) -> str:  # pylint: disable=no-self-use
        """
        Pre-processing function used to register the corresponding parameter as the mode for the current experiment. The
        mode is a string that can be used to differentiate between different experiments with the same config. Usual
        modes can include 'train', 'test', 'eval', 'debug'... It can also be a list of strings if several successive
        modes are to be used. It does not actually do anything, the only effect is the registration, so that the mode
        can then be accessed using "config[config.get_hook('mode')[0]]".
        Priority : INDIFFERENT (0)
        YAML tag : mode

        :param mode: string or list of strings corresponding to the mode(s)
        :return: the same string or list of strings
        """
        return mode

    @hook("tracker_config")
    @assign_order(Priority.INDIFFERENT)  # does not depend on or change parameter value, only processes parameter name
    @assign_yaml_tag("tracker_config", "pre", "dict")
    def register_as_tracker_config(self, tracker_config: dict) -> dict:  # pylint: disable=no-self-use
        """
        Pre-processing function used to register the corresponding parameter as the tracker config. The tracker config
        is a dict that contains at least one key : 'type'. Valid types are given by the 'ACCEPTED_TRACKERS' variable in
        experiment.py and refer to the type of tracker used. Other keys in the dict depend on the parameters required by
        corresponding tracker type.
        Priority : INDIFFERENT (0)
        YAML tag : tracker_config

        :param tracker_config: dict corresponding to the tracker config
        :raises: ValueError: if the tracker config is not a dict or does not contain at least a parameter called 'type'
        which is a string
        :raises: ValueError: if the 'type' is not recognised
        :raises: ValueError: if there are missing required keys for this 'type'
        :return: the same dict
        """
        if os.getenv('NODE_RANK'):  # do not track if in a pytorch-lightning spawned process
            return {"type": []}
        required_keys = {
            "basic": [],
            "sacred": ["db_url", "db_name"],
            "mlflow": ["tracking_uri"],
            "tensorboard": ["logdir"],
            "clearml": ["project_name"],
        }
        possible_keys = list(required_keys.values()) + [["basic_logdir", "sub_loggers", "type"]]
        if not isinstance(tracker_config, dict):
            raise ValueError(f"{tracker_config} is not a valid tracker config : it is not a dict.")
        if "type" not in tracker_config:
            raise ValueError(f"{tracker_config} is not a valid tracker config : it has no 'type' key.")
        if tracker_config["type"] is not None and not isinstance(tracker_config["type"], str) and not (
                isinstance(tracker_config["type"], list) and all(isinstance(i, str) for i in tracker_config["type"])):
            raise ValueError(f"{tracker_config} is not a valid tracker config : 'type' should be None, a string or a"
                             " list of strings.")
        if isinstance(tracker_config["type"], str):
            tracker_list = [t.strip(" ") for t in tracker_config["type"].split(",")]
        elif tracker_config["type"] is not None:
            tracker_list = []
            for string in tracker_config["type"]:
                tracker_list += [t.strip(" ") for t in string.split(",")]
        else:
            tracker_list = []
        if any(tracker not in required_keys for tracker in tracker_list):
            raise ValueError(f"Unknown tracker among {tracker_list}. "
                             f"Accepted values are {list(required_keys.keys())}.")
        for tracker in tracker_list:
            for key in required_keys[tracker]:
                if key not in tracker_config:
                    raise ValueError(f"Missing key in {tracker}-type tracker config : '{key}'.")
        for key in tracker_config:
            if not any(key in key_list for key_list in possible_keys):
                YAECS_LOGGER.warning(f"WARNING : Unknown key '{key}' in tracker config. It might be ignored.")
        tracker_config["type"] = tracker_list
        return tracker_config



"""Configuration file for the Sphinx documentation builder."""
# pylint: disable=all
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import sys

from setuptools_scm import get_version

sys.path.insert(0, os.path.abspath('../'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'YAECS'
copyright = '2022, Reactive Reality AG'  # noqa A001
author = 'Reactive Reality AG'
try:
    release = get_version()
except:  # noqa E722
    release = get_version(root='..', relative_to=__file__)

# Remove modules docstring


def remove_module_docstring(app, what, name, obj, options, lines):
    if what == "module":
        del lines[:]


def setup(app):
    app.connect("autodoc-process-docstring", remove_module_docstring)

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
]

autodoc_mock_imports = [
    "pytorch_lightning",
    "tqdm",
]

master_doc = "index"
autoapi_type = 'python'
autoapi_dirs = ["yaecs"]

autodoc_default_options = {
    "member-order": "bysource",
    "undoc-members": True,
}

add_module_names = False
autoclass_content = "both"
napoleon_use_param = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
}

templates_path = ['_templates']
exclude_patterns = ['_build']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']

html_theme_options = {
    "canonical_url": "",
    "analytics_id": "UA-XXXXXXX-1",
    "logo_only": False,
    "display_version": True,
    "prev_next_buttons_location": "both",
    "style_external_links": "#ff9900",
    "style_nav_header_background": "#ff9900",
    # Toc options
    "collapse_navigation": False,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}
html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "valentingol",  # Username
    "github_repo": "LoggerML",  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}



"""
Reactive Reality Machine Learning Config System - Usage example
(various utility functions and classes)
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Lesser Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import os

import numpy as np
from scipy import stats


class Metrics:
    """Metrics class utilities."""

    @staticmethod
    def exponential_l2(ground_truth, predictions):
        return ((np.exp(ground_truth) - predictions)**2).mean()

    @staticmethod
    def square_root_l2(ground_truth, predictions):
        return ((np.sqrt(1 + ground_truth) - predictions)**2).mean()


def create_data(data_config):
    mu = data_config.generation.mean  # pylint: disable=invalid-name
    sigma = data_config.generation.variance
    train = stats.truncnorm((-1-mu) / sigma, (1-mu) / sigma, loc=mu,
                            scale=sigma).rvs(data_config.train.size)
    test = stats.truncnorm((-1-mu) / sigma, (1-mu) / sigma, loc=mu,
                           scale=sigma).rvs(data_config.test.size)
    return train, test


def create_and_train_model(model_config, train_samples):

    def model(x):
        if model_config.type == "linear":
            return (1 + x * model_config.param1 + x * model_config.param2
                    + x * model_config.param3)
        if model_config.type == "polynomial":
            return (1 + x * model_config.param1
                    + np.power(x, 2) * model_config.param2
                    + np.power(x, 3) * model_config.param3)
        return None

    # Added train samples for reference, but they're not actually used here
    _ = train_samples
    return model


def test_model_and_return_metrics(model, test_samples, metrics):
    predictions = model(test_samples)
    return {
        metric: getattr(Metrics, metrics[metric])(test_samples, predictions)
        for metric in metrics
    }


def log_experiment(config, metrics):
    exp_dir = os.path.join("log", config.name)
    if not os.path.exists(exp_dir):
        os.makedirs(exp_dir)
    run_dir = os.path.join(exp_dir, f"run_{len(os.listdir(exp_dir))}")
    os.makedirs(run_dir)
    config.save(os.path.join(run_dir, "config_save.yaml"))
    with open(os.path.join(run_dir, "metrics.json"), 'w',
              encoding='utf-8') as fil:
        json.dump(metrics, fil)



""" This module contains classes to track and manage time durations. """

import logging
import time
from typing import Dict, List, Optional, Tuple, Union

YAECS_LOGGER = logging.getLogger(__name__)


class Timer:
    """ This class is a timer. It has a name to know what it records, and start and stop times. """
    def __init__(self, name: str = "MyTimer", verbose: Optional[int] = 1, start: bool = False,
                 step: Optional[int] = None):
        """
        Creates a timer.

        :param name: name of the timer
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        :param start: whether to start the timer immediately
        :param step: starting step if start is True (if None, assumes step=0, timings are averages over elapsed steps)
        """
        self.name: str = name
        self.start_times: List[Tuple[Union[float, int]]] = []
        self.stop_times: List[Tuple[Union[float, int]]] = []
        self.verbose: int = 1 if verbose is None else verbose
        if start:
            self.start(step=step)

    def __str__(self) -> str:
        return self.render("current") if self.running else self.render("last")

    def __getitem__(self, item: Union[int, str]) -> Optional[float]:
        return self.get(which=item, step_aggregation="average")

    @property
    def empty(self) -> bool:
        """ Returns whether the timer has completed any run. """
        return len(self.stop_times) == 0

    @property
    def running(self) -> bool:
        """ Returns whether the timer is currently running. """
        return len(self.start_times) > len(self.stop_times)

    @property
    def verbose(self) -> int:
        """ Returns the verbosity level of the timer. """
        return self._verbose

    @verbose.setter
    def verbose(self, verbose: int) -> None:
        """ Sets the verbosity level of the timer. """
        if verbose not in [0, 1, 2]:
            raise ValueError(f"Invalid value for 'verbose' : {verbose}. Can be 0, 1 or 2.")
        self._verbose = verbose

    def get(self, which: Union[int, str] = "last", step_aggregation: str = "average") -> Optional[float]:
        """
        Gets a specified recorded timing (the last one by default). A recorded timing needs a start and a stop.

        :param which: which timing to get. Can be an int (index of the timing in the list of recorded timings), or
            'last' (last recorded timing), 'first' (first recorded timing), 'average' (average of all recorded timings),
            'total' (sum of all recorded timings) or 'current' (current timing)
        :param step_aggregation: whether to return each timing as an average over the steps it was recorded on or as a
            total. Can be 'average' or 'total'
        :raises ValueError: if which is not an int or in ['last', 'first', 'average', 'total']
        :raises IndexError: if which is an int and is out of range
        :return: the timing
        """
        current_time = time.time()

        def _aggregate(value, number_of_steps):
            return value / number_of_steps if step_aggregation == "average" else value
        if step_aggregation not in ["average", "total"]:
            raise ValueError(f"Invalid value for 'step_aggregation' : {step_aggregation}. Can be 'average' or 'total'.")
        which = self._process_which(which, possible_strings=["last", "first", "average", "total", "current"])

        if which == "current":
            return current_time - self.start_times[-1][0] if self.running else None

        if self.empty:
            return 0 if which == "total" else None

        if which == "average":
            step_divider = sum(self.get_number_of_steps(i) for i in range(len(self.stop_times)))
            divider = step_divider if step_aggregation == "total" else len(self.stop_times)
            return sum(_aggregate(self.stop_times[i][0] - self.start_times[i][0], self.get_number_of_steps(i))
                       for i in range(len(self.stop_times))) / divider

        if which == "total":
            return sum(_aggregate(self.stop_times[i][0] - self.start_times[i][0], self.get_number_of_steps(i))
                       for i in range(len(self.stop_times)))

        return _aggregate(self.stop_times[which][0] - self.start_times[which][0], self.get_number_of_steps(which))

    def get_at_step(self, step: Optional[Union[int, str]] = None) -> Optional[float]:
        """
        Gets the duration of given step, or None if given step was not recorded. Similar to 'get' except it takes steps
        instead of timing indices. Also accepts 'last' and 'first' as steps.

        :param step: if None last step, else step to get
        :return: the duration if it was recorded, otherwise None
        """
        if isinstance(step, str):
            return self.get(which=step, step_aggregation="average")
        index = None
        for i, stop_time in enumerate(self.stop_times):
            if stop_time[1] > step:
                index = i
                break
        if self.start_times[index][1] > step or index is None:
            return None
        return self.get(which=index, step_aggregation="average")

    def get_number_of_steps(self, which: Union[int, str] = "last") -> Optional[int]:
        """
        Gets the number of steps of a specified recorded timing (the last one by default).

        :param which: which timing to get. Can be an int (index of the timing in the list of recorded timings), or
            'last' (last recorded timing), 'first' (first recorded timing), 'average' (average of all recorded timings)
            or 'total' (sum of all recorded timings)
        :return: the number of steps
        """
        which = self._process_which(which, possible_strings=["last", "first", "average", "total"])
        if self.empty:
            return 0 if which == "total" else None
        if which in ["average", "total"]:
            return self.stop_times[-1][1] - self.start_times[0][1]
        return self.stop_times[which][1] - self.start_times[which][1]

    def render(self, which: Union[int, str] = "current", step_aggregation: str = "average",
               verbose: Optional[int] = None) -> str:
        """
        Renders a string to display all or part of the timer's internal state.

        :param which: which timing to get. Can be an int (index of the timing in the list of recorded timings), or
            'current' (current timing), 'last' (last recorded timing), 'first' (first recorded timing), 'average'
            (average of all recorded timings), 'total' (sum of all recorded timings) or 'all' (all recorded timings)
        :param step_aggregation: whether to return each timing as an average over the steps it was recorded on or as a
            total. Can be 'average' or 'total'
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        :raises ValueError: if verbose is not in [0, 1, 2]
        :return: the rendered string
        """
        def _render_duration(which, step_aggregation, verbose):
            rendered = ""
            duration = self.get(which=which, step_aggregation=step_aggregation)
            if duration is None:
                if verbose == 2:
                    rendered += "no timing recorded so far"
                if verbose == 1:
                    rendered += "none"
                if verbose == 0:
                    rendered += "-"
            else:
                duration = format_duration(duration, human_readable=bool(verbose))
                number_of_steps = self.get_number_of_steps(which)
                if verbose == 2:
                    timing_name = f"{which} timing" if isinstance(which, str) else f"timing {which}"
                    step_operation = "averaged" if step_aggregation == "average" else "accumulated"
                    step_rendering = f" ({step_operation} over {number_of_steps} steps)" if number_of_steps > 1 else ""
                    rendered += f"{timing_name} : {duration}{step_rendering}"
                if verbose == 1:
                    step_operation = "/" if step_aggregation == "average" else "in"
                    step_rendering = f" ({step_operation} {number_of_steps} steps)" if number_of_steps > 1 else ""
                    rendered += f"{duration}{step_rendering}"
                if verbose == 0:
                    rendered += f"{duration}"
            return rendered
        current_time = time.time()
        verbose = self.verbose if verbose is None else verbose
        if verbose not in [0, 1, 2]:
            raise ValueError(f"Invalid value for 'verbose' : {verbose}. Can be 0, 1 or 2.")
        which_tmp = self._process_which(which, possible_strings=["current", "last", "first", "average", "total", "all"])
        which = which if which in ["last", "first"] else which_tmp
        rendered_string = ""

        # Set header
        if verbose == 2:
            rendered_string += f"---- '{self.name}' timer ----\n"
            header_size = len(rendered_string)
        if verbose == 1:
            rendered_string += f"'{self.name}' : "
        if verbose == 0:
            rendered_string += f"{self.name}("

        # Set summary
        if verbose == 2:
            steps = sum(self.stop_times[i][1] - self.start_times[i][1] for i in range(len(self.stop_times)))
            rendered_string += f"Overall {len(self.stop_times)} recorded timings totalling {steps} steps.\n"

        # Set body
        if which == "current":
            if self.running:
                duration = format_duration(current_time - self.start_times[-1][0], human_readable=bool(verbose))
                if verbose == 2:
                    rendered_string += f"Currently running for {duration}.\n"
                if verbose == 1:
                    rendered_string += f"running for {duration}"
            else:
                duration = "-"
                if verbose == 2:
                    rendered_string += "Currently not running.\n"
                if verbose == 1:
                    rendered_string += "not running"
            if verbose == 0:
                rendered_string += f"{duration}"
        elif which != "all":
            duration = _render_duration(which, step_aggregation, verbose)
            if verbose == 2:
                rendered_string += f"{duration.capitalize()}.\n"
            else:
                rendered_string += duration
        else:
            if self.empty:
                if verbose == 2:
                    rendered_string += "No timing recorded so far.\n"
                if verbose == 1:
                    rendered_string += "none"
                if verbose == 0:
                    rendered_string += "-"
            else:
                if verbose == 2:
                    rendered_string += "All recorded timings :\n - "
                    rendered_string += " ;\n - ".join([_render_duration(i, step_aggregation, verbose)
                                                       for i in range(len(self.stop_times))])
                    rendered_string += ".\n"
                else:
                    rendered_string += " ; ".join([_render_duration(i, step_aggregation, verbose)
                                                  for i in range(len(self.stop_times))])

        # Set footer
        if verbose == 2:
            rendered_string += "-" * header_size
        if verbose == 1:
            rendered_string += "."
        if verbose == 0:
            rendered_string += ")"

        return rendered_string

    def reset(self) -> None:
        """ Resets the timer, removing all recorded timings. """
        self.start_times = []
        self.stop_times = []

    def start(self, step: Optional[int] = None) -> float:
        """
        Automatically stops any previously started timer, and starts a new one.

        :param step: starting step (if none, assumes step=last stop step, timings are averages over the elapsed steps)
        :raises ValueError: if step is less than the previous stopping step
        :return: the starting timestamp
        """
        current_time = time.time()
        self.stop(step=step, _time=current_time)
        previous_step = 0
        if not self.empty:
            previous_step = self.stop_times[-1][1]
        step = previous_step if step is None else step
        if step < previous_step:
            raise ValueError(f"Invalid value for 'step' : {step}. Must be greater than or equal to the previous "
                             "stopping step.")
        self.start_times.append((current_time, step))
        if self.verbose == 2:
            YAECS_LOGGER.info(f"Timer '{self.name}' started at step {step} at time "
                              f"{time.asctime(time.localtime(current_time))}.")
        return current_time

    def stop(self, step: Optional[int] = None, _time: Optional[float] = None) -> Optional[float]:
        """
        Stops any previously started timer, or does nothing if no timer is running.

        :param step: starting step (if none assumes step=starting step + 1, timings are averages over the elapsed steps)
        :raises ValueError: if step is less than or equal to the previous starting step
        :return: the duration of the timer
        """
        current_time = time.time() if _time is None else _time
        if not self.running:
            return None
        previous_step = self.start_times[-1][1]
        step = previous_step + 1 if step is None else step
        if step <= previous_step:
            raise ValueError(f"Invalid value for 'step' : {step}. Must be greater than the previous starting step.")
        self.stop_times.append((current_time, step))
        ellapsed = current_time - self.start_times[-1][0]
        if self.verbose == 2:
            YAECS_LOGGER.info(f"Timer '{self.name}' stopped at step {step} : ellapsed time "
                              f"{format_duration(ellapsed)} (on average "
                              f"{format_duration(ellapsed/(step-previous_step))} per step).")
        return ellapsed

    def _process_which(self, which: Union[int, str], possible_strings: List[str]) -> Union[int, str]:
        """
        Processes 'which' arguments.

        :raises ValueError: if which is not an int or in possible_strings
        :raises IndexError: if which is an int and is out of range
        :return: the processed which
        """
        if isinstance(which, int) and which >= len(self.stop_times):
            raise IndexError(f"Index {which} is out of range for timer {self.name} with {len(self.stop_times)} "
                             "recorded timings.")
        if not isinstance(which, int) and which not in possible_strings:
            raise ValueError(f"Invalid value for 'which' : {which}. Can be an int or in {possible_strings}.")
        if which == "last":
            which = len(self.stop_times) - 1
        elif which == "first":
            which = 0
        return which


class TimerManager:
    """ This class records timings and prints or returns them in a variety of formats. """
    def __init__(self, verbose: Optional[int] = None):
        """
        Creates a timer manager use to start and stop timers, as well as render them.

        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail. If None, set based on local logger
        """
        self.timers: Dict[str, Timer] = {}
        level = YAECS_LOGGER.getEffectiveLevel()
        level_verbose = 0 if level > 30 else 1 if level > 10 else 2
        verbose = level_verbose if verbose is None else verbose
        self.verbose: int = verbose

    def __getitem__(self, item: Union[int, str]) -> Dict[str, Optional[float]]:
        item = self._process_which(item)
        if item == "current":
            return {name: timer.get("current") for name, timer in self.timers.items()}
        if item == "average":
            return {name: timer.get("average", step_aggregation="total") for name, timer in self.timers.items()}
        if item == "total":
            return {name: timer.get("total", step_aggregation="total") for name, timer in self.timers.items()}
        return {name: timer.get_at_step(item) for name, timer in self.timers.items()}

    @property
    def first_step(self) -> int:
        """ Gets the first step of the timer manager (assuming it is the first step of all its timers). """
        return min(timer.start_times[0][1] for timer in self.timers.values())

    @property
    def last_step(self) -> int:
        """ Gets the last or current step of the timer manager (assuming it is the latest step of all its timers). """
        return max(timer.start_times[-1][1] for timer in self.timers.values())

    def render(self, which_step: Union[int, str] = "current", verbose: Optional[int] = None) -> str:
        """
        Renders a string to display some properties of the timers.

        :param which_step: which timing to get from the timers. Can be an int (index of the step), or 'current'
            (currently running timers), 'last' (last recorded step), 'first' (first recorded step),
            'average' (average of all recorded steps) or 'total' (sum of all recorded steps)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail. If None, uses each timer's verbose
        :return: the rendered string
        """
        rendered_string = ""
        which_step = self._process_which(which_step)
        durations = self[which_step]
        verbose = self.verbose if verbose is None else verbose

        # Set header
        if which_step == "current":
            which_string = "(currently) "
        elif which_step == "average":
            which_string = f"(averaged over {self.last_step - self.first_step} steps) "
        elif which_step == "total":
            which_string = f"(accumulated over {self.last_step - self.first_step} steps) "
        else:
            which_string = f"(at step {which_step}) "
        if verbose == 2:
            rendered_string += f"\n---- Reporting on {len(self.timers)} timers {which_string}----\n"
            header_size = len(rendered_string)
        if verbose == 1:
            rendered_string += f"\n{len(self.timers)} timers {which_string}: \n"
        if verbose == 0:
            rendered_string += "{"

        # Set body
        if not durations:
            if verbose in [1, 2]:
                rendered_string += "No timer set so far.\n"
            if verbose == 0:
                rendered_string += " - "
        for name, duration in durations.items():
            if duration is None:
                continue
            if verbose in [1, 2]:
                rendered_string += f" - {name} : {format_duration(duration)}\n"
            if verbose == 0:
                rendered_string += f"{name}:{format_duration(duration)} ; "
                if name == list(durations.keys())[-1]:
                    rendered_string = rendered_string[:-3]

        # Set footer
        if verbose in [1, 2]:
            not_reported = [name for name, duration in durations.items() if duration is None]
            if len(not_reported) > 0:
                reported_string = "running" if which_step == "current" else "recorded "
                reported_string += "so far" if which_step in ["total", "average"] else ""
                reported_string += f"at step {which_step}" if isinstance(which_step, int) else ""
                rendered_string += f"{len(not_reported)} timers not {reported_string} ({', '.join(not_reported)}).\n"
        if verbose == 2:
            rendered_string += "-" * header_size
        if verbose == 0:
            rendered_string += "}"

        return rendered_string

    def reset(self) -> None:
        """ Resets the manager, deleting all timers. """
        self.timers = {}

    def start(self, name: str = "MyTimer", step: Optional[int] = None, verbose: Optional[int] = None) -> None:
        """
        Starts a timer.

        :param name: name of the timer
        :param step: starting step (if None, assumes step=last stop step, timings are averages over the elapsed steps)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        try:
            with TemporaryVerbose(self.timers[name], verbose):
                self.timers[name].start(step=step)
        except KeyError:
            self.timers[name] = Timer(name=name, start=True, step=step, verbose=verbose)

    def stop(self, name: str = "MyTimer", step: Optional[int] = None, verbose: Optional[int] = None) -> None:
        """
        Stops a timer.

        :param name: name of the timer
        :param step: starting step (if None, assumes step=start step + 1, timings are averages over the elapsed steps)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        with TemporaryVerbose(self.timers[name], verbose):
            self.timers[name].stop(step=step)

    def update(self, start: Union[None, str, List[str]] = None, stop: Union[None, str, List[str]] = None,
               step: Optional[int] = None, verbose: Optional[int] = None) -> None:
        """
        Automatically starts and stops timers.

        :param start: names of the timers to start
        :param stop: names of the timers to stop
        :param step: starting step (if None, assumes step=last stop step, timings are averages over the elapsed steps)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        if start is None:
            start = []
        if isinstance(start, str):
            start = [start]
        if stop is None:
            stop = []
        if isinstance(stop, str):
            stop = [stop]
        for name in stop:
            self.stop(name=name, step=step, verbose=verbose)
        for name in start:
            self.start(name=name, step=step, verbose=verbose)

    def _process_which(self, which: Union[int, str]) -> Union[int, str]:
        """
        Processes 'which' arguments.

        :raises ValueError: if which is not an int or in ['current', 'last', 'first', 'average', 'total']
        :raises IndexError: if which is an int and is out of range
        :return: the processed which
        """
        if isinstance(which, int) and which > self.last_step:
            raise IndexError(f"Unknown step {which} : the last recorded timing in on step {self.last_step}.")
        possible_strings = ["current", "last", "first", "average", "total"]
        if not isinstance(which, int) and which not in possible_strings:
            raise ValueError(f"Invalid value for 'which' : {which}. Can be an int or in {possible_strings}.")
        if which == "last":
            which = self.last_step
        elif which == "first":
            which = self.first_step
        return which


class TemporaryVerbose:
    """ Context manager to change the verbosity of a timer. """
    def __init__(self, timer: Timer, verbose: int):
        """
        Creates a context manager to change the verbosity of a timer.

        :param timer: timer to change the verbosity of
        :param verbose: new verbosity level
        """
        self.timer: Timer = timer
        self.verbose: int = verbose
        self.old_verbose: Optional[int] = None

    def __enter__(self) -> None:
        """ Changes the verbosity level of the timer. """
        if self.verbose is not None:
            self.old_verbose = self.timer.verbose
            self.timer.verbose = self.verbose

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """ Resets the verbosity level of the timer. """
        if self.old_verbose is not None:
            self.timer.verbose = self.old_verbose


class TimeInContext:
    """ This Context starts a Timer in given TimerManager at the start of the context and stops it at the end. """
    def __init__(self, timer_manager: TimerManager, name: str = "MyTimer", step: Optional[int] = None,
                 verbose: Optional[int] = None):
        """
        Creates a context to start a timer in a timer manager.

        :param timer_manager: timer manager to start the timer in
        :param name: name of the timer
        :param step: starting step (if None, assumes step=last stop step, timings are averages over the elapsed steps)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        self.timer_manager: TimerManager = timer_manager
        self.name: str = name
        self.step: Optional[int] = step
        self.verbose: Optional[int] = verbose

    def __enter__(self) -> None:
        """ Starts the timer in the timer manager. """
        self.timer_manager.start(name=self.name, step=self.step, verbose=self.verbose)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """ Stops the timer in the timer manager. """
        self.timer_manager.stop(name=self.name, verbose=self.verbose)


def format_duration(duration: float, precision: int = 2, human_readable: bool = True) -> str:
    """ Formats a duration in seconds to a string. """
    if not human_readable:
        return f"{duration:.{precision}f}"
    groups = {"w": 604800, "d": 86400, "h": 3600, "m": 60}
    formatted = ""
    for name, seconds in groups.items():
        if duration >= seconds:
            formatted += f"{int(duration // seconds)}{name} "
            duration %= seconds
    return f"{formatted}{duration:.{precision}f}s"



""" This file defines a library of convenient processing functions. """

import logging
import os
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, List, Optional

from ..yaecs_utils import (Priority,
                           assign_order, assign_yaml_tag, compare_string_pattern, set_function_attribute)

if TYPE_CHECKING:
    from numbers import Number

    from .config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class ConfigProcessingFunctionsMixin:
    """ Pre- and Post-processing functions Mixin class for YAECS configurations. """

    get_main_config: Callable[[], 'Configuration']
    get_processed_param_name: Callable[[bool], str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # ----- PRE-PROCESSING -----

    def check_param_in_list(self, list_of_choices: List[Any]) -> Callable:
        """
        Returns a pre-processing function that checks if a param value belongs to a list. Returned function has
        order OFTEN_FIRST (-10).
        For example in your pre-processing dict there could be a line such as this :
        "mode": self.check_param_in_list(["train", "val", "test", "infer"])

        :param list_of_choices: list of valid parameter values
        :return: checking function
        """

        def _check(param: Any, choices: List[Any], config: 'Configuration') -> Any:
            if param is None:
                return param
            if param.lower() not in choices:
                possibilities = ""
                for index, choice in range(len(choices)):
                    possibilities += f"'{choice}'"
                    if index != len(choices) - 1:
                        possibilities += ", "
                param_name = config.get_processed_param_name(full_path=True)
                raise ValueError(f"Invalid value for param '{param_name}': '{param}'."
                                 "Valid choices are [{possibilities}].")
            return param

        return_function = partial(_check, choices=list_of_choices, config=self)
        set_function_attribute(return_function, "__name__", "check_param_in_list")
        set_function_attribute(return_function, "order", Priority.OFTEN_FIRST)
        return return_function

    @assign_order(Priority.ALWAYS_LAST)  # there would most likely not be any other processing for this param
    @assign_yaml_tag("copy", "pre", "str")
    def copy_param(self, path_to_copy: str) -> str:
        """
        This pre-processing function declares a param as being an exact copy of another param. Its default value must
        be name of the param it copies. It will be protected against modifications by further config merges (only the
        param being copied can be modified), but can still pe post-processed further after its copy by post-processing
        functions with an order above OFTEN_LAST (10).
        Priority : ALWAYS_LAST (20)
        YAML tag : copy

        :param path_to_copy: path of param from which to copy the value
        :return: path_to_copy
        """

        def _copy(param: str, main: 'Configuration') -> Any:
            params = main.match_params(param)
            if len(params) == 1:
                return main[params[0]]
            if len(params) > 1:
                raise RuntimeError(f"Ambiguous replacement : param name '{param}' was matched to multiple "
                                   f"params : '{params}'.")
            return param

        if not isinstance(path_to_copy, str):
            raise TypeError(f"The default value of params declared as copies of other params must be the full path to "
                            f"the copied param. Current default value '{path_to_copy}' is not a string.")
        main = self.get_main_config()
        param_name = self.get_processed_param_name(full_path=True)
        copy_fn = partial(_copy, main=main)
        set_function_attribute(copy_fn, "__name__", "_copy")
        set_function_attribute(copy_fn, "order", Priority.OFTEN_LAST)

        current_processing = object.__getattribute__(main, "_added_post_processing")()
        if not any((compare_string_pattern(param_name, k) and v.__name__ == "_copy")
                   for k, v in current_processing.items()):
            main.add_processing_function_all(param_name, copy_fn, "post")

        return self.protected_param(path_to_copy)

    def check_number_in_range(self, minimum: 'Number' = -float('inf'), maximum: 'Number' = float('inf')) -> Callable:
        """
        Returns a pre-processing function that checks if a numerical param value is in a range. Returned function has
        order OFTEN_FIRST (-10).
        For example in your pre-processing dict there could be a line such as this :
        "probability": self.number_in_range(minimum=0, maximum=1)

        :param minimum: minimal valid value for the parameter
        :param maximum: maximal valid value for the parameter
        :return: checking function
        """

        def _check(param: 'Number', minimum_: 'Number', maximum_: 'Number', config: 'Configuration') -> 'Number':
            if param is None:
                return param
            if not minimum_ <= param <= maximum_:
                param_name = config.get_processed_param_name(full_path=True)
                raise ValueError(f"Invalid value for param '{param_name}': '{param}'. "
                                 f"Must be in range [{minimum_} ; {maximum_}].")
            return param
        return_function = partial(_check, minimum_=minimum, maximum_=maximum, config=self)
        set_function_attribute(return_function, "__name__", "check_number_in_range")
        set_function_attribute(return_function, "order", Priority.OFTEN_FIRST)
        return return_function

    @assign_order(Priority.ALWAYS_LAST)  # there can never be any subsequent processing for this param
    @assign_yaml_tag("protected", "pre", "Any")
    def protected_param(self, param: Any) -> Any:
        """
        This pre-processing function declares a param as being protected against modifications. This means that only the
        default config can change its value, and any attempt to set it from another config will raise an error.
        Priority : ALWAYS_LAST (20)
        YAML tag : protected

        :param param: value of the param to protect
        :return: param
        """
        main = self.get_main_config()
        param_name = self.get_processed_param_name(full_path=True)
        prev = main.match_params(param_name)
        if prev:
            raise RuntimeError(f"Parameter '{param_name}' is protected, it cannot be set from configs other than the "
                               f"default config.")
        return param

    # ----- POST-PROCESSING -----

    @assign_order(Priority.OFTEN_LAST)  # should happen after register_as_experiment_path
    @assign_yaml_tag("sub_folder", "post", "Optional[str]")
    def folder_in_experiment(self, folder: Optional[str]) -> Optional[str]:
        """
        Returns a post-processing function that extends a path assuming it is located in the experiment path, then the
        corresponding folder is created. Requires an experiment_folder to have been declared.
        Priority : OFTEN_LAST (10)
        YAML tag : sub_folder

        :param folder: path of the subfolder within the experiment folder
        :return: checking function
        """
        experiment_path = self.get_main_config().get_experiment_path()
        path = os.path.join(experiment_path, folder).rstrip(os.path.sep)
        os.makedirs(path, exist_ok=True)
        return path

    def folder_in_experiment_if(self, condition_list: List[tuple] = None) -> Callable:
        """
        Returns a post-processing function that extends a path assuming it is located in the experiment path. Then, if
        the conditions listed in condition_list ar all met, the corresponding folder is created. Returned function has
        order OFTEN_LAST (10).
        condition_list is a list of conditions represented by tuples of length 2 or 3 :

        * when a condition is represented by a tuple of length 2, the first element should be the path to parameters in
          the main config and the second element should be a value for those parameters. The condition will return true
          if the corresponding parameters have the corresponding value ;
        * when a condition is represented by a tuple of length 3, the behaviour is the same except the third element is
          a function to apply to the parameters values before they are compared to the second element.
          For example in your post-processing dict there could be a line such as this :
          "save_weights_path": self.folder_in_experiment(condition_list=[("mode", "train")])

        :param condition_list: list of conditions to be met for the path to be created
        :return: checking function
        """

        def _folder_in_experiment(folder: str, config: 'Configuration', conditions: list) -> str:
            experiment_path = config.get_experiment_path()
            path = os.path.join(experiment_path, folder).rstrip(os.path.sep)
            conditions = {str(nb): [c for c in conditions if len(c) == nb] for nb in [2, 3]}
            if (all(all(c[2](config[i]) == c[1] for i in config.match_params(c[0])) for c in conditions["3"])
                    and all(all(config[i] == c[1] for i in config.match_params(c[0])) for c in conditions["2"])):
                os.makedirs(path, exist_ok=True)
            return path

        if condition_list is None:
            condition_list = []
        if not isinstance(condition_list, list):
            raise TypeError(f"Argument 'condition_list' of function 'folder_in_experiment' should be a list. "
                            f"Got {type(condition_list)}.")
        if not all(isinstance(c, tuple) for c in condition_list):
            raise TypeError("All elements of argument 'condition_list' of function 'folder_in_experiment' should be "
                            "tuples.")
        if not all(len(c) == 2 or len(c) == 3 for c in condition_list):
            raise ValueError("All elements of argument 'condition_list' of function 'folder_in_experiment' should have "
                             "2 elements (param and value) or 3 elements (param, value and conversion).")
        return_function = partial(_folder_in_experiment, config=self.get_main_config(), conditions=condition_list)
        set_function_attribute(return_function, "__name__", "folder_in_experiment_if")
        set_function_attribute(return_function, "order", Priority.OFTEN_LAST)
        return return_function



""" This file defines the class for the ClearML logger. """
import importlib.util
import logging
import os
from typing import Any, Optional, Union

from .base_logger import Logger
from .logger_utils import lazy_import, value_to_float
from ..experiment_utils import format_mode

# ClearML integration, see original package : https://clear.ml
clearml = lazy_import("clearml")  # pylint: disable=invalid-name

YAECS_LOGGER = logging.getLogger(__name__)


class ClearMLLogger(Logger):
    """ ClearML Logger. This logger logs things to a ClearML server setup in a clearml.conf config. """
    def __init__(self, tracker):
        super().__init__("ClearML Logger", tracker)
        self.task = None
        self.project_name = self.tracker.config.get("project_name", None)

    def check_config_requirements(self) -> str:
        if self.project_name is None:
            return ("Config Requirement Error : no project name was provided for the clearml logger. Provide one using "
                    "the project_name key in the tracker config.")
        return ""

    def check_install(self) -> str:
        error = ""
        if not os.path.isfile(os.path.expanduser("~/clearml.conf")):
            error = ("Install Error : no 'clearml.conf' file detected in your home directory. Please run 'clearml-init'"
                     " as explained in https://clear.ml/docs/latest/docs/getting_started/ds/ds_first_steps.")
        if not importlib.util.find_spec("clearml"):
            error += "\n" if error else ""
            error += "Install Error : your experiment tracking config requires clearml - currently not installed !"
        return error

    def get_logger_object(self) -> Any:
        return self.task

    def main_function_context(self):
        return CMLContext(self.tracker)

    def start_run(self, experiment_name: str, run_name: str, description: str, params: dict) -> None:
        self.task = clearml.Task.init(project_name=self.project_name,
                                      task_name=f"{experiment_name}/{run_name}",
                                      task_type=self._get_tast_type(),
                                      continue_last_task=bool(os.getenv("PICKUP")))
        self.task.set_comment(description)
        self.task.connect(params)

    def log_scalar(self, name: str, value: Union[float, int], step: Optional[int] = None,
                   sub_logger: Optional[str] = None, description: Optional[str] = None) -> None:
        value = value_to_float(value, self.name)
        if value == "":
            return

        if step is None:
            self.task.logger.report_single_value(name=name, value=value)
        else:
            if sub_logger is None:
                names_hierarchy = name.split("/")
                if len(names_hierarchy) > 1:
                    title = "/".join(names_hierarchy[:-1])
                else:
                    title = names_hierarchy[-1]
                series = names_hierarchy[-1]
            else:
                title = name
                series = sub_logger
            self.task.logger.report_scalar(title=title, series=series, value=value, iteration=step)
        if description is not None:
            YAECS_LOGGER.warning("WARNING : in log_scalar : 'description' is not used in clearml.")

    def log_image(self, name: str, image, step: Optional[int] = None, sub_logger: Optional[str] = None) -> None:
        # Support for matplotlib figures logging, see original package : https://matplotlib.org/
        matplotlib = lazy_import("matplotlib")  # pylint: disable=invalid-name
        # Support to save numpy arrays as images, see original package : https://numpy.org/
        np = lazy_import("numpy")  # pylint: disable=invalid-name
        # Support to save numpy arrays as images, see original package : https://pypi.org/project/Pillow/
        Image = lazy_import("PIL.Image")  # pylint: disable=invalid-name
        # Support for plotly figures, see original package : https://plotly.com/python/
        plotly = lazy_import("plotly")  # pylint: disable=invalid-name
        if isinstance(image, str):
            self.task.logger.report_image(title=name, series=sub_logger, local_path=image, iteration=step)
        elif isinstance(image, np.ndarray):
            self.task.logger.report_image(title=name, series=sub_logger, image=image.astype(np.uint8), iteration=step)
        elif isinstance(image, Image):
            self.task.logger.report_image(title=name, series=sub_logger, image=np.array(image, dtype=np.uint8),
                                          iteration=step)
        elif isinstance(image, matplotlib.figure.Figure):
            self.task.logger.report_matplotlib(title=name, series=sub_logger, iteration=step, figure=image)
        elif isinstance(image, plotly.graph_objects.Figure):
            self.task.logger.report_plotly(title=name, series=sub_logger, iteration=step, figure=image)
        else:
            self.task.logger.report_image(title=name, series=sub_logger, image=np.array(image, dtype=np.uint8),
                                          iteration=step)

    def _get_tast_type(self):
        """ Returns the task type for the ClearML task, inferred from the experiment mode if there is one. """
        task_types = {
            "TRAINING": clearml.Task.TaskTypes.training,
            "TESTING": clearml.Task.TaskTypes.testing,
            "VALIDATION": clearml.Task.TaskTypes.testing,
            "INFERENCE": clearml.Task.TaskTypes.inference,
            "DATA PROCESSING": clearml.Task.TaskTypes.data_processing,
            "DEBUG": clearml.Task.TaskTypes.training,
        }
        mode_param = self.tracker.experiment.config.get_hook("mode")
        mode = format_mode(self.tracker.experiment.config[mode_param[0]]) if mode_param else "TRAINING"
        task_type = clearml.Task.TaskTypes.custom
        for key, value in task_types.items():
            if key in mode:
                task_type = value
                break
        return task_type


class CMLContext:
    """ Class used to set up the context for ClearML's tracker """

    def __init__(self, tracker):
        """
        Initialises a context used to close the ClearML runs when they are done.

        :param tracker: tracker object where to find the ClearML runs
        """
        self.tracker = tracker

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tracker.loggers["clearml"].close()



""" This file defines the experiment tracker object, which for an experiment handles all tracking and logging
activities. """
import logging
import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from .timer import TimerManager, TimeInContext
from .loggers.aggregate_logger import AggregateLogger
from ..config.config import Configuration
from ..yaecs_utils import compare_string_pattern, NoValue

YAECS_LOGGER = logging.getLogger(__name__)


class Tracker:
    """ Class created by Experiment to log values. """

    def __init__(self, tracker_config: Dict[str, Any], experiment, experiment_name: Optional[str] = None,
                 run_name: Optional[str] = None, starting_step: int = 0, params_filter_fn: Optional[Callable] = None,
                 log_modified_params_only: bool = False, do_not_log_hooks: bool = False, 
                 only_params_to_log: Optional[List[str]] = None, params_not_to_log: Optional[List[str]] = None):
        """
        Reads the tracker config from the general config to create a Tracker object used for logging during the run.

        :param tracker_config: tracker config from the general config
        :param experiment: passed automatically from the instance of Experiment this tracker originates from
        :param experiment_name: name for the experiment (inferred from the experiment path if not provided)
        :param run_name: name for the run (inferred from the experiment path if not provided)
        :param starting_step: step at which to start logging scalars
        :param params_filter_fn: function to use instead of the default filter to get the list of the names of the
            parameters to log to the tracker from the config. If this is used, then 'log_modified_params_only',
            'only_params_to_log' and 'params_not_to_log' are ignored.
        :param log_modified_params_only: whether the parameters to filter using the other arguments are the parameters
            that changed compared to the default config (True) or only those of the whole config (False)
        :param do_not_log_hooks: whether the parameters defined as hooks should be filtered out (True) or not (False)
        :param only_params_to_log: if provided, only the parameters whose names are given will be filtered and logged
        :param params_not_to_log: if provided, parameters whose names are given will be filtered out
        """
        # General attributes
        self.config = {k: v for k, v in tracker_config.items() if k != "type"}
        loggers = self.config.get("sub_loggers", [])
        self.sub_loggers = [""] + (loggers if isinstance(loggers, list) else [loggers])
        self.experiment = experiment
        self.experiment_name = experiment_name
        self.run_name = run_name
        self._step = starting_step
        self.timer = TimerManager()

        # AggregateLogger
        self.loggers = AggregateLogger(self, tracker_config.get("type", []))
        self.loggers.check_install()
        self.loggers.check_config_requirements()

        # Parameters filtering
        self.get_filtered_params = self.default_filter if params_filter_fn is None else params_filter_fn
        self.log_modified_params_only = log_modified_params_only
        self.do_not_log_hooks = do_not_log_hooks
        self.only_params = None
        if only_params_to_log is not None:
            if params_filter_fn is not None:
                YAECS_LOGGER.warning("WARNING : parameter 'only_params_to_log' is ignored when 'params_filter_fn' is "
                                     "provided.")
            self.only_params = only_params_to_log if isinstance(only_params_to_log, list) else [only_params_to_log]
        self.except_params = []
        if params_not_to_log is not None:
            if params_filter_fn is not None:
                YAECS_LOGGER.warning("WARNING : parameter 'params_not_to_log' is ignored when 'params_filter_fn' is "
                                     "provided.")
            self.except_params = params_not_to_log if isinstance(params_not_to_log, list) else [params_not_to_log]

    def default_filter(self, config: Configuration) -> List[str]:
        """
        Default parameters filtering function. Uses the pre-defined 'log_modified_params_only', 'only_params' and
        'except_params' attributes to filter out the parameters of the config.
        It starts with a list of all the params which were modified from the default config, except if
        log_modified_params_only was set to False in which case it starts from all the parameters in the config. Then,
        it filters out all hooks, then all parameters in except_params, and finally only keeps those that are also in
        only_params.

        :param config: instance of Configuration from which to filter the parameters
        :return: the list of the filtered parameters names
        """
        # Get params to filter
        if self.log_modified_params_only:
            default = config.__class__.build_from_configs(config.config_metadata["config_hierarchy"][0],
                                                          do_not_post_process=True, do_not_merge_command_line=True,
                                                          verbose=False)
            names = [config.match_params("*" + difference[0]) for difference in default.compare(config, reduce=True)]
            if any(len(name) > 1 for name in names):
                raise RuntimeError("ERROR : Compared parameter resolved to 2 parameter names.")
            names = [name[0] for name in names]
        else:
            names = config.get_parameter_names(deep=True, no_sub_config=True)

        # Get filters
        filters = list(self.except_params)
        if self.do_not_log_hooks:
            for hooks in config.get_hook().values():
                filters += hooks

        # Perform filtering
        to_return = []
        for name in names:
            if self.only_params is None or any(
                    compare_string_pattern(config.match_params(name)[0], pattern)
                    for pattern in self.only_params):
                if not any(
                        compare_string_pattern(config.match_params(name)[0], pattern) for pattern in filters):
                    to_return.append(name)
        return to_return

    def extract_names(self) -> Tuple[str, str]:
        """
        Gets experiment and run names, using either names given when instantiating the Experiment (if provided) or
        inferring them from the config's experiment path. If nothing is provided and no experiment path is defined,
        raises an error.

        :raises RuntimeError: if no experiment path is defined in the config and no experiment or run name is provided
        :return: the experiment name and the run name
        """
        try:
            name_from_config = self.experiment.config.get_experiment_path()
        except RuntimeError:
            name_from_config = None
        if name_from_config is None and self.experiment_name is None:
            raise RuntimeError("Please specify an experiment name either by "
                               "registering a parameter as experiment path "
                               "with the 'register_as_experiment_path' method, "
                               "or by passing it when creating your experiment object.")
        if name_from_config is None and self.run_name is None:
            raise RuntimeError("Please specify a run name either by "
                               "registering a parameter as experiment path "
                               "with the 'register_as_experiment_path' method, "
                               "or by passing it when creating your experiment object.")
        if self.experiment_name is None:
            if self.experiment.config.get_variation_name() is None:
                exp_name = "_".join(name_from_config.split(os.sep)[-1].split("_")[:-1])
            else:
                exp_name = "_".join(name_from_config.split(os.sep)[-2].split("_")[:-1])
        else:
            exp_name = self.experiment_name
        if self.run_name is None:
            if self.experiment.config.get_variation_name() is None:
                run_name = name_from_config.split(os.sep)[-1].split("_")[-1]
            else:
                run_name = name_from_config.split(os.sep)[-2].split("_")[-1] + "_" + name_from_config.split(os.sep)[-1]
        else:
            run_name = self.run_name
        return exp_name, run_name

    def start_run(self, description: Optional[str] = None) -> None:
        """
        Initialises the configured loggers, which most of the time means preparing their logger in self.loggers.
        """
        experiment_name, run_name = self.extract_names()
        config = self.experiment.config
        params_to_log = {}
        for pattern in self.get_filtered_params(config):
            param_name = config.match_params(pattern)[0]
            params_to_log[pattern] = config.get_pre_post_processing_values().get(param_name, config[param_name])
        self.loggers.start_run(experiment_name, run_name, description, params_to_log)

    def step(self, step: Optional[int] = None, auto_log_timers: bool = True, print_timers: bool = True) -> None:
        """
        Increments the step counter.

        :param step: step to increment to. If None, will increment to the next step
        :param auto_log_timers: whether to automatically log the timers at the end of the step
        :param print_timers: whether to print the timers at the end of the step
        """
        if step is not None and step < self._step:
            raise ValueError(f"Cannot go back to step {step} from step {self._step}.")
        if auto_log_timers:
            timers = {"timers/" + name: duration for name, duration in self.timer["last"].items()
                      if duration is not None}
            self.log_scalars(timers, step=self._step)
        if print_timers:
            print(self.timer.render(which_step="last"))
        self._step = self._step + 1 if step is None else step

    def log_image(self, name: str, image, step: Optional[int] = None, sub_logger: Optional[str] = None,
                  maximum: Optional[int] = None, maximum_per_step: Optional[int] = None,
                  main_process_only: bool = False) -> None:
        """
        Logs an image using the logger. The image could be a path to a saved image, matplotlib or plotly figure, a
        PIL.Image, or a n*n*3 numpy array.

        :param name: name for the value to be logged
        :param image: image to be logged. Accepted formats are strings (interpreted as paths to saved images),
            matplotlib or plotly figures, PIL images, or n*n*3 numpy arrays
        :param step: step at which the value is logged. If set to None or a negative value, will default to 0 for the
            tensorboard tracker, will default to -1 for the basic tracker and will be logged as a "single value" for the
            clearml tracker. If not provided, will default to the current step of the tracker (0 by default)
        :param sub_logger: if specified, logs to corresponding sub-logger. Can be interpreted as a sub-folder for the
            scalar name most of the time, but in the case of tensorboard will actually use a different summary writer
        :param maximum: maximum number of images to log. The logger stops logging images once this number is reached.
            If None or negative value, no maximum is set.
        :param maximum_per_step: maximum number of images to log per step. The logger stops logging images once this
            number is reached for a given step. If None or negative value, no maximum is set.
        :param main_process_only: do not try to log in pytorch-lightning sub-processes
        """
        step = self._step if isinstance(step, NoValue) else step
        if not main_process_only or not os.getenv('NODE_RANK'):  # do not track in a pytorch-lightning spawned process
            self._warn_if_no_logs()
            self.loggers.log_image(name=name, image=image, step=step, sub_logger=sub_logger, maximum=maximum,
                                   maximum_per_step=maximum_per_step)

    def log_scalar(self, name: str, value: Union[float, int], step: Union[NoValue, None, int] = NoValue(),
                   sub_logger: Optional[str] = None, description: Optional[str] = None,
                   main_process_only: bool = False) -> None:
        """
        Logs the given value under the given name at given step in the configured trackers. The description is optional
        and can only be used if the tracker is tensorboard.

        :param name: name for the value to be logged
        :param value: value to be logged
        :param step: step at which the value is logged. If set to None or a negative value, will default to 0 for the
            tensorboard tracker, will default to -1 for the basic tracker and will be logged as a "single value" for the
            clearml tracker. If not provided, will default to the current step of the tracker (0 by default)
        :param sub_logger: if specified, logs to corresponding sub-logger. Can be interpreted as a sub-folder for the
            scalar name most of the time, but in the case of tensorboard will actually use a different summary writer
        :param description: only used for the tensorboard tracker, corresponds to a short description of the value
        :param main_process_only: do not try to log in pytorch-lightning sub-processes
        """
        step = self._step if isinstance(step, NoValue) else step
        if not main_process_only or not os.getenv('NODE_RANK'):  # do not track in a pytorch-lightning spawned process
            self._warn_if_no_logs()
            self.loggers.log_scalar(name, value, step=step, sub_logger=sub_logger, description=description)

    def log_scalars(self, dictionary: Dict[str, Any], step: Union[NoValue, None, int] = NoValue(),
                    sub_logger: Optional[str] = None, main_process_only: bool = False) -> None:
        """
        Logs several values contained in a dictionary, one by one using Tracker.log_scalar.

        :param dictionary: dictionary containing the (name, value) pairs to be logged
        :param step: step at which the value is logged. If not provided, will default to 0 for the tensorboard tracker,
            will default to -1 for the basic tracker and will be logged as a "single value" for the clearml tracker
        :param sub_logger: if specified, logs to corresponding sub-logger. Can be interpreted as a sub-folder for the
            scalar name most of the time, but in the case of tensorboard will actually use a different summary writer
        :param main_process_only: do not try to log in pytorch-lightning sub-processes
        """
        if not main_process_only or not os.getenv('NODE_RANK'):  # do not track in a pytorch-lightning spawned process
            self._warn_if_no_logs()
            for key, value in dictionary.items():
                self.log_scalar(key, value, step=step, sub_logger=sub_logger)

    def start_timer(self, name: str = "MyTimer", step: Union[NoValue, None, int] = NoValue(),
                    verbose: Optional[int] = None) -> None:
        """
        Starts a timer.

        :param name: name of the timer
        :param step: Step at which to log the measured duration. If not provided, uses the internal step. If None,
            assumes step=previous stop step.
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        step = self._step if isinstance(step, NoValue) else step
        if name in self.timer.timers and self.timer.timers[name].running:
            self.stop_timer(name, step, verbose)
        self.timer.start(name, step, verbose)

    def stop_timer(self, name: str = "MyTimer", step: Union[NoValue, None, int] = NoValue(),
                   verbose: Optional[int] = None) -> None:
        """
        Stops a timer.

        :param name: name of the timer
        :param step: Step at which to log the measured duration. If not provided, uses the internal step. If None,
            assumes step=previous start step + 1.
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        step = self._step if isinstance(step, NoValue) else step
        if name in self.timer.timers and self.timer.timers[name].running:
            start_step = self.timer.timers[name].start_times[-1][1]
            step = step if step > start_step else start_step + 1
        self.stop_timer(name, step, verbose)

    def measure_time(self, name: str = "MyTimer", step: Union[NoValue, None, int] = NoValue(),
                     verbose: Optional[int] = None) -> None:
        """
        Returns an instanciated context manager to time a block of code.

        :param name: name of the timer
        :param step: Step at which to log the measured duration. If not provided, uses the internal step. If None,
            assumes step=previous stop step
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        :return: context manager to time a block of code
        """
        step = self._step if isinstance(step, NoValue) else step
        return TimeInContext(timer_manager=self.timer, name=name, step=step, verbose=verbose)

    def _warn_if_no_logs(self):
        if not self.loggers.has_loggers:
            YAECS_LOGGER.warning("WARNING : no tracker configured, scalars will not be logged anywhere.")
            if os.getenv('NODE_RANK'):
                YAECS_LOGGER.warning("This is because trackers are deactivated in pytorch-lightning processes.\n"
                                     "To suppress this message, pass 'main_process_only=True'.")



""" Defines common functions for logger objects. """
from bisect import bisect
import importlib.util
import io
import logging
import os
import sys
from types import ModuleType
from typing import Any

YAECS_LOGGER = logging.getLogger(__name__)


def add_to_csv(csv_path: str, name: str, value: Any, step: int) -> None:
    """
    Adds a logged value to the csv containing previously logged values

    :param csv_path: path to the csv containing the logged values
    :param name: name of the value to log
    :param value: value of the value to log
    :param step: step for which to log the value
    """
    if os.path.isfile(csv_path):
        with open(csv_path, encoding='utf-8') as csv_file:
            data = csv_file.readlines()
            steps = [int(d.split(",")[0]) for d in data[1:]]
            metrics = [d.strip("\n") for d in data[0].split(",")[1:]]
            values = [[d.split(",")[1 + i].strip("\n") if d.split(",")[1 + i] else ""
                       for d in data[1:]] for i in range(len(metrics))]
    else:
        steps = []
        metrics = []
        values = []
    if name not in metrics:
        metrics.append(name)
        values.append(["" for _ in range(len(steps))])
    if step not in steps:
        index = bisect(steps, step)
        steps.insert(index, step)
        for i, metric in enumerate(metrics):
            values[i].insert(index, str(value) if metric == name else "")
    else:
        values[metrics.index(name)][steps.index(step)] = str(value)

    with open(csv_path, 'w', encoding='utf-8') as csv_file:
        csv_file.write(",".join(["steps"] + metrics) + "\n")
        for index, step_to_log in enumerate(steps):
            data = [str(step_to_log)] + [v[index] for v in values]
            csv_file.write(",".join(data) + "\n")


def lazy_import(name: str) -> ModuleType:
    """
    Imports a module in such a way that it is only loaded in memory when it is actually used.
    Implementation from https://docs.python.org/3/library/importlib.html#implementing-lazy-imports.

    :param name: name of the module to load
    :return: the loaded module
    """
    spec = importlib.util.find_spec(name)
    if not spec:
        return None
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module


def new_print(*args, sep: str = " ", end: str = "", file: io.TextIOWrapper = None, **keywords) -> None:
    """
    Replaces the builtin print function during an experiment run such that printed messages are also logged. Please note
    that the default file (None) logs to logging's root logger which will always go to the next line after each message.
    Therefore, the 'end' param does not replace \\n as usual, but adds a suffix after the message and before the \\n.

    :param args: objects to print
    :param sep: how to separate the different objects
    :param end: suffix to add after the message
    :param file: file to print to, defaults to a logging to logging's root logger with level logging.INFO
    :param keywords: might contain 'flush', in which case raise an error
    :raises TypeError: when the keyword arguments contain 'flush'
    """
    if not os.getenv('NODE_RANK'):  # do not print if in a pytorch-lightning spawned process
        if "flush" in keywords:
            raise TypeError("Because YAECS uses logging.info to log messages logged via the print function, the 'flush'"
                            " parameter is not supported for the print function within your main.")
        message = sep.join([str(a) for a in args]) + end.strip()
        if message.strip():
            if file is not None and file is not sys.stdout:
                file.write(message)
            logging.getLogger("yaecs.print_catcher").info(message)


def value_to_float(value: Any, logger_name: str) -> str:
    """ Converts a value to a float if possible, otherwise raises a warning and returns an empty string. """
    try:
        value = float(value)
    except ValueError:
        YAECS_LOGGER.warning(f"WARNING : will not log non-float value {value} to {logger_name} logger.")
        return ""
    return value



"""
Reactive Reality Machine Learning Config System - _ConfigurationBase object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import copy
import logging
import os
import sys
from collections.abc import Iterable
from functools import partial
from numbers import Real
from pathlib import Path
from typing import (TYPE_CHECKING,
                    Any, Callable, Dict, List, Optional, Tuple, Type, Union)

import yaml

from ..yaecs_utils import (YAML_EXPRESSIONS,
                           ConfigDeclarator, TypeHint,
                           compare_string_pattern, compose, format_str, get_quasi_bash_sys_argv, get_order,
                           is_type_valid, parse_type, recursive_set_attribute, set_function_attribute, update_state)
from .config_convenience import ConfigConvenienceMixin
from .config_getters import ConfigGettersMixin
from .config_hooks import ConfigHooksMixin
from .config_processing_functions import ConfigProcessingFunctionsMixin
from .config_setters import ConfigSettersMixin

if TYPE_CHECKING:
    from .config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class _ConfigurationBase(ConfigHooksMixin, ConfigGettersMixin, ConfigSettersMixin, ConfigConvenienceMixin,
                         ConfigProcessingFunctionsMixin):
    """ Base class for YAECS configurations. Defines its basic behaviour, such as creation and merging operations,
    including processing- and type checking-related logic, but not its constructors (see Configuration class for those,
    and its docstring for more details about the composition of the Configuration superclass). """

    add_processing_function: Callable[[str, Callable, str], None]
    add_processing_function_all: Callable[[str, Callable, str], None]
    config_metadata: dict
    parameters_pre_processing: Callable[[], Dict[str, Callable]]
    parameters_post_processing: Callable[[], Dict[str, Callable]]
    _get_instance: Callable
    _get_tagged_methods_info: Callable[[], List[Tuple[Union[str, Callable]]]]
    _main_config: 'Configuration'
    _methods: List[str]
    _nesting_hierarchy: List[str]
    _operating_creation_or_merging: bool
    _protected_attributes: List[str]
    _state: List[str]
    _verbose: bool

    def __init__(self, from_argv: str = "", do_not_pre_process: bool = False, do_not_post_process: bool = False):
        """
        Should never be called directly by the user. Please use one of the constructors defined for the Configuration
        class instead, or the utils.make_config convenience function.

        :param from_argv: pattern used to find the config in the command line arguments, or "" if not applicable
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        :raises ValueError: if the overwriting regime is not valid
        :return: none
        """

        # PROTECTED ATTRIBUTES
        self._assigned_as_yaml_tags = {processor[0]: (processor[3], processor[1], processor[2])
                                       for processor in self._get_tagged_methods_info()}
        self._former_saving_time = None
        self._from_argv = from_argv
        self._modified_buffer = []
        self._post_process_master_switch = not do_not_post_process
        self._pre_process_master_switch = not do_not_pre_process
        for process_type in ["pre", "post"]:
            setattr(self, f"_{process_type}_processing_functions", {})
            self._prepare_processing_functions(process_type)
        self._pre_postprocessing_values = {}
        self._reference_folder = None
        self._sub_configs_list = []
        self._type_hints = {}
        self._was_last_saved_as = None
        super().__init__()

    def __getitem__(self, item) -> Any:
        if "." in item and "*" not in item:
            sub_config_name = ("___"
                               + item.split(".")[0] if item.split(".")[0] in self._methods else item.split(".")[0])
            sub_config = getattr(self, sub_config_name)
            if not isinstance(sub_config, _ConfigurationBase):
                did_you_mean_message = self._did_you_mean(sub_config_name, filter_type=self.__class__)
                raise TypeError(f"As the parameter '{sub_config_name}' is not a sub-config"
                                f", it cannot be accessed.\n{did_you_mean_message}")
            return sub_config[item.split(".", 1)[1]]
        return getattr(self, "___" + item if item in self._methods else item)

    def __setattr__(self, key, value) -> None:
        if (self.is_in_operation() or self._main_config.is_in_operation()
                or self.config_metadata["overwriting_regime"] == "unsafe"):
            object.__setattr__(self, key, value)
        elif self.config_metadata["overwriting_regime"] == "auto-save":
            self._manual_merge({key: value}, source='code')
        elif self.config_metadata["overwriting_regime"] == "locked":
            raise RuntimeError("Overwriting params in locked configs "
                               "is not allowed.")
        else:
            raise ValueError(f"No behaviour determined for value '"
                             f"{self.config_metadata['overwriting_regime']}' of "
                             f"parameter 'overwriting_regime'.")

    def __getattribute__(self, item) -> Any:
        try:
            return object.__getattribute__(self, item)
        except AttributeError as exception:
            if not item.startswith("_") and not any(state.startswith("setup") for state in self._state):
                raise AttributeError(f"Unknown parameter of the configuration : '{item}'.\n"
                                     f"{self._did_you_mean(item)}") from exception
            raise AttributeError from exception

    def __iter__(self):
        return iter(self._get_user_defined_attributes())

    def check_type(self, type_or_types: TypeHint) -> Callable:
        """
        Returns a processing function that checks for given type. Can be used for example with the following line in a
        parameters post-processing dict:
        "parameter_that_should_be_int": self.check_type(int)

        * The type can be any of None, bool, int, float, str, dict, list. The value 0 instead means no type check.
        * Unions are denoted by tuples of types.
        * You can specify the type of the elements of your lists by using a list of types. This list should contain
          either one type (in which case the list is expected to only contain elements of that type) or as many types as
          there are elements in the list (in which case each element is tested with the corresponding type)
        * You can specify the type of the elements of your dicts by using a dict or a set of types. If you use a set, it
          can only contain one type (in which case the dict is expected to contain only values of that type).
          If you use a dict of types, the keys used in that dict that match the keys in the parameter will be checked
          using the values as types.

        :param type_or_types: type for which to create the function
        :return: the processing function
        """
        def _check_type(value: Any, type_to_check: TypeHint, original_type: TypeHint) -> Any:
            def _wrong_type() -> None:
                name = self.get_processed_param_name(full_path=False)
                is_full = original_type == type_to_check
                checked_type = type(type_to_check) if isinstance(type_to_check, (list, dict, set)) else type_to_check
                raise ValueError(f"{'Parameter' if is_full else 'Part of parameter'} '{name}' (value : {value})\n"
                                 f"has incorrect type '{type(value)}'. Expected '{checked_type}'.")

            if isinstance(type_to_check, tuple):
                if not type_to_check:
                    raise ValueError("Undefined behaviour for empty tuples. Maybe you meant to use an empty list or "
                                     "dict ?")
                fails = True
                for to_check in type_to_check:
                    try:
                        _check_type(value, to_check, original_type)
                    except ValueError:
                        pass
                    else:
                        fails = False
                if fails:
                    _wrong_type()

            elif isinstance(type_to_check, list):
                if not isinstance(value, list):
                    _wrong_type()
                if len(type_to_check) > 1:
                    if len(type_to_check) != len(value):
                        raise ValueError("When providing a list of types, its length must be one or match the length of"
                                         " the value.")
                    for v_to_check, t_to_check in zip(value, type_to_check):
                        _check_type(v_to_check, t_to_check, original_type)
                else:
                    types = type_to_check[0] if type_to_check else 0
                    for i in value:
                        _check_type(i, types, original_type)

            elif isinstance(type_to_check, dict):
                if not isinstance(value, dict):
                    _wrong_type()
                if not type_to_check:
                    raise ValueError("Undefined behaviour for empty dicts. Maybe you meant to use an empty list or "
                                     "{\"type\": ...} ?")
                if len(type_to_check) > 1:
                    raise ValueError("When providing a dict of types, its length must be 1. Maybe you meant to use a"
                                     " tuple ?")
                for i in value:
                    _check_type(value[i], type_to_check[list(type_to_check.keys())[0]], original_type)

            elif type_to_check != 0 and type_to_check is not None and not isinstance(value, type_to_check):
                if not (type_to_check is float and isinstance(value, int)):
                    _wrong_type()

            elif type_to_check is None and value is not None:
                _wrong_type()
            return value

        return partial(_check_type, type_to_check=type_or_types, original_type=type_or_types)

    @update_state("init_from_config;_name")
    def init_from_config(self, config_path_or_dict: ConfigDeclarator) -> None:
        """
        Entrypoint for all methods trying to get any value from outside the config to inside the config. This
        includes creating new parameters when creating the config or merging existing parameters after the creation.
        Users should only use this to merge or create parameters during a creation or merge operation (for instance in a
        processing function). If you want to merge parameters in your main code, outside a constructor or other
        operation, please use self.merge.

        :param config_path_or_dict: path or dictionary for the config to merge
        """
        if config_path_or_dict is not None:
            if isinstance(config_path_or_dict, str):
                with open(self._find_path(config_path_or_dict), encoding='utf-8') as yaml_file:
                    for dictionary_to_add in yaml.load_all(yaml_file, Loader=self._get_yaml_loader()):
                        for item in self._superficial_dict_cleanup(dictionary_to_add).items():
                            self._process_item_to_merge_or_add(item)
            else:
                for item in config_path_or_dict.items():
                    self._process_item_to_merge_or_add(item)

    def merge(self, config_path_or_dictionary: ConfigDeclarator, do_not_pre_process: bool = False,
              do_not_post_process: bool = False) -> None:
        """
        Merges provided config path of dictionary into the current config.

        :param config_path_or_dictionary: path or dictionary for the config to merge
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        """
        self._manual_merge(config_path_or_dictionary=config_path_or_dictionary, do_not_pre_process=do_not_pre_process,
                           do_not_post_process=do_not_post_process)

    def merge_from_command_line(self, to_merge: Optional[Union[List[str], str]] = None,
                                do_not_pre_process: bool = False, do_not_post_process: bool = False) -> None:
        """
        Formerly used to manually merge the command line arguments into the config, which is now done automatically and
        thus should no longer be done manually. Can still be used to manually merge a string emulating command line
        arguments.

        :param to_merge: if specified, merges this string or list of strings instead of the sys.argv list of strings
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        """
        if self._verbose and to_merge is None:
            # TODO handle all warnings with a function _warn that logs and stores messages for future prints
            YAECS_LOGGER.warning("WARNING : merge_from_command_line is now deprecated and will automatically start "
                                 "after using any constructor.\nYou can remove the 'config.merge_from_command_line()' "
                                 "line from your code now :) it's redundant.")
        to_merge = self._gather_command_line_dict(to_merge)
        if to_merge:
            self._manual_merge(to_merge, do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process,
                               source='command line')

    @staticmethod
    def _added_pre_processing():
        """ Will contain pre-processing function added via self.add_processing_function_all. """
        return {}

    @staticmethod
    def _added_post_processing():
        """ Will contain post-processing function added via self.add_processing_function_all. """
        return {}

    def _check_for_unlinked_sub_configs(self) -> None:
        """ Used to raise an error when unlinked sub-configs are declared. """
        all_configs = self.get_all_sub_configs()
        linked_configs = self.get_all_linked_sub_configs()
        for i in all_configs:
            found_correspondence = False
            for j in linked_configs:
                if self._are_same_sub_configs(i, j):
                    found_correspondence = True
                    break
            if not found_correspondence:
                raise RuntimeError(f"Sub-config '{i.get_name()}' is unlinked. Unlinked "
                                   "sub-configs are not allowed.")

    def _find_path(self, path: str) -> str:
        """ Used to find a config from its (potentially relative) path, because it might be ambiguous relative to where
        it should be looked for. Probably very improvable. """
        def _get_path(path_to_check):
            if os.path.exists(path_to_check):
                return os.path.abspath(path_to_check)
            if path_to_check.endswith(".yaml") or path_to_check.endswith(".yml"):
                to_check = ".".join(path_to_check.split(".")[:-1])
            else:
                to_check = path_to_check
            there = [os.path.exists(to_check + ".yaml"), os.path.exists(to_check + ".yml"), os.path.exists(to_check)]
            if sum(there) == 0:
                return None
            if sum(there) == 1:
                return os.path.abspath([to_check + ".yaml", to_check + ".yml", to_check][there.index(True)])
            there.pop(there.index(False))
            raise RuntimeError(f"Ambiguity for provided path '{path_to_check}' : detected two possible paths :\n"
                               f"   - {there[0]}\n   - {there[1]}")

        # If the path is absolute, use it...
        if os.path.isabs(path):
            path = _get_path(path)
            if path is not None:
                self._reference_folder = str(Path(path).parents[0])
                return path

        # ... if not, search relatively to some reference folders.
        else:
            possibilities = []
            last_is_current = False

            # First check relatively to parent configs' directories...
            for config in reversed(self.config_metadata["config_hierarchy"]):
                if isinstance(config, str):
                    relative_path = os.path.join(Path(config).parents[0], path)
                    absolute_path = _get_path(relative_path)
                    if absolute_path is not None and absolute_path not in possibilities:
                        possibilities.append(absolute_path)

            # ... then also check the current reference folder since
            # the config hierarchy is not always up-to-date...
            if self._reference_folder is not None:
                relative_path = os.path.join(self._reference_folder, path)
                absolute_path = _get_path(relative_path)
                if absolute_path is not None and absolute_path not in possibilities:
                    possibilities.append(absolute_path)
            if self._main_config is not None and self._main_config.get_reference_folder() is not None:
                relative_path = os.path.join(self._main_config.get_reference_folder(), path)
                absolute_path = _get_path(relative_path)
                if absolute_path is not None and absolute_path not in possibilities:
                    possibilities.append(absolute_path)

            # ... and finally, check relatively to the current
            # working directory.
            absolute_path = _get_path(path)
            if absolute_path is not None and absolute_path not in possibilities:
                last_is_current = True
                possibilities.append(absolute_path)

            if len(possibilities) > 1:
                if path.endswith(".yaml"):
                    filtered = [p for p in possibilities if p.endswith(".yaml")]
                elif path.endswith(".yml"):
                    filtered = [p for p in possibilities if p.endswith(".yml")]
                else:
                    filtered = [p for p in possibilities if not p.endswith(".yaml") and not p.endswith(".yml")]
            else:
                filtered = possibilities

            if len(filtered) > 1:
                YAECS_LOGGER.warning(f"WARNING : Multiple matches for path {path}. '{filtered[0]}' will be used.\n"
                                     f"All matches : {filtered}.")
            if filtered:
                if last_is_current and filtered[0] == possibilities[-1]:
                    self._reference_folder = str(Path(absolute_path).parents[0])
                return filtered[0]

        raise FileNotFoundError(f"ERROR : path not found ({path}).")

    def _get_yaml_loader(self) -> Type[yaml.FullLoader]:
        """ Used to get a custom YAML loader capable of parsing config tags. """

        def generic_constructor(yaml_loader, tag, node):

            is_param_tag = bool(yaml_loader.constructed_objects)

            # If type tag, handle type-hinting or processing assignment.
            if tag[1:].lower().startswith("type:") and tag[1:].lower() != "type:config":
                if not is_param_tag:
                    raise RuntimeError(f"'{tag[1:]}' is not a valid sub-config name.")
                if not any(state.startswith("setup") for state in self._state):
                    raise RuntimeError("Type-hinting is only allowed in the default config.")
                name = yaml_loader.constructed_objects[list(yaml_loader.constructed_objects.keys())[-1]]
                name = self._get_full_path(name)
                type_hint = tag[6:]
                if type_hint in self._assigned_as_yaml_tags:
                    _, processor_type, new_type_hint = self._assigned_as_yaml_tags[type_hint]
                    if self.get_variation_name() is None:
                        self.add_processing_function_all(name, f"_tagged_method_{type_hint}", processor_type)
                    type_hint = new_type_hint
                self._main_config.add_type_hint(name, parse_type(type_hint))
                if isinstance(node, yaml.ScalarNode):
                    if node.value == "":
                        def _can_be_str(parsed_type):
                            if parsed_type is str:
                                return True
                            if isinstance(parsed_type, tuple):
                                if str in parsed_type:
                                    return True
                                return any(_can_be_str(t) for t in parsed_type)
                            return False
                        if _can_be_str(parse_type(tag[6:])):
                            return yaml_loader.yaml_constructors["tag:yaml.org,2002:str"](yaml_loader, node)
                    for key, value in YAML_EXPRESSIONS.items():
                        if value.match(node.value):
                            return yaml_loader.yaml_constructors[f"tag:yaml.org,2002:{key}"](yaml_loader, node)
                    return yaml_loader.construct_scalar(node)
                if isinstance(node, yaml.SequenceNode):
                    return yaml_loader.construct_sequence(node, deep=True)
                if isinstance(node, yaml.MappingNode):
                    return yaml_loader.construct_mapping(node, deep=True)

            # Otherwise, assume sub-config tag
            if is_param_tag:
                # Case 1 : not root of the YAML file > infer name from parameter name
                sub_configs_names = list(yaml_loader.constructed_objects.keys())[-1].value.split(".")
                sub_config = sub_configs_names[0]
                apply_to_node = ".".join(sub_configs_names[1:])
                if tag[1:].lower() != "type:config":
                    YAECS_LOGGER.warning(f"WARNING : Naming sub-configs is deprecated. Tag '{tag}' will be ignored and "
                                         "should be removed. They might be used for a different purpose in a future "
                                         "release."
                                         "\nSince now YAML dicts are implicitly considered sub-configs, be sure to tag "
                                         "with '!type:dict' if you want to use a dict.")
            else:
                if tag[1:].lower() == "type:config":
                    # Case 2 : name not provided and root of the YAML file > assume dict
                    return yaml_loader.construct_mapping(node, deep=True)
                # Case 3 : name provided and root of the YAML file > use first part as config name
                sub_configs_names = tag[1:].split(".")
                sub_config = sub_configs_names[0]
                apply_to_node = ".".join(sub_configs_names[1:])

            if apply_to_node:
                # Add prefix to all sub-config parameters
                for i in range(len(node.value)):  # pylint: disable=consider-using-enumerate
                    node.value[i][0].value = apply_to_node + "." + node.value[i][0].value
            self._nesting_hierarchy.append(sub_config)  # needs to be set temporarily (potential recursive call)
            to_convert = yaml_loader.construct_mapping(node)
            new_config = self._get_instance(name=sub_config,
                                            config_path_or_dictionary=self._superficial_dict_cleanup(to_convert),
                                            nesting_hierarchy=self._nesting_hierarchy, state=self._state,
                                            main_config=self._main_config, verbose=self._verbose)
            self.set_sub_config(new_config)
            self._nesting_hierarchy.pop(-1)
            return new_config if is_param_tag else {sub_config: new_config}

        loader = yaml.FullLoader
        loader.DEFAULT_MAPPING_TAG = "!type:config"
        yaml.add_multi_constructor("", generic_constructor, Loader=loader)
        return loader

    def _manual_merge(self, config_path_or_dictionary: ConfigDeclarator, do_not_pre_process: bool = False,
                      do_not_post_process: bool = False, source: str = 'config',
                      ) -> None:
        """ This method is called whenever a merge is done by the user, and not by the config creation process. It
        simply calls _merge with some additional bookkeeping. """
        self._merge(config_path_or_dictionary=config_path_or_dictionary, do_not_pre_process=do_not_pre_process,
                    do_not_post_process=do_not_post_process, source=source)
        self._post_process_modified_parameters()
        self.set_post_processing(True)
        if self.get_main_config().config_metadata["overwriting_regime"] == "auto-save":
            if self.get_main_config().get_save_file() is not None:
                self.get_main_config().save()

    @update_state("merging;_name")
    def _merge(self, config_path_or_dictionary: ConfigDeclarator, do_not_pre_process: bool = False,
               do_not_post_process: bool = False, source: str = 'config') -> None:
        """ Method handling all merging operations to call init_from_config with the proper bookkeeping. """
        if self._main_config == self:
            object.__setattr__(self, "_operating_creation_or_merging", True)
            if self._verbose:
                YAECS_LOGGER.info(f"Merging from {source} : {format_str(config_path_or_dictionary)}")
            self.set_post_processing(not do_not_post_process)
            self.set_pre_processing(not do_not_pre_process)
            self.init_from_config(config_path_or_dictionary)
            self.config_metadata["config_hierarchy"].append(config_path_or_dictionary)
            self._check_for_unlinked_sub_configs()
            self.set_pre_processing(True)
            self._operating_creation_or_merging = False
        else:
            dicts_to_merge = []
            if isinstance(config_path_or_dictionary, str):
                with open(self._find_path(config_path_or_dictionary), encoding='utf-8') as yaml_file:
                    for dictionary_to_add in yaml.load_all(yaml_file, Loader=self._get_yaml_loader()):
                        dicts_to_merge.append(self._superficial_dict_cleanup(dictionary_to_add))
                    yaml_file.close()
            else:
                dicts_to_merge.append(config_path_or_dictionary)
            for dictionary in dicts_to_merge:
                self._main_config._merge(  # pylint: disable=protected-access
                    {self._get_full_path(a): b
                     for a, b in dictionary.items()}, do_not_pre_process=do_not_pre_process,
                    do_not_post_process=do_not_post_process, source=source,
                )

    @update_state("working_on;_name")
    def _process_item_to_merge_or_add(self, item: Tuple[str, Any]) -> None:
        """ Method called by init_from_config to merge or add a given key, value pair. """
        key, value = item

        # Process metadata. If there is metadata, treat the rest of
        # the merge as "loading a saved file"... (which will deactivate
        # the parameter pre-processing for this merge)
        if key == "config_metadata":
            pattern = "Saving time : * (*) ; Regime : *"
            if not isinstance(value, str) or not compare_string_pattern(value, pattern):
                raise RuntimeError("'config_metadata' is a special parameter. "
                                   "Please do not edit or set it.")

            regime = value.split(" : ")[-1]
            if regime == "unsafe" and self._verbose:
                YAECS_LOGGER.warning("WARNING : YOU ARE LOADING AN UNSAFE CONFIG FILE. Reproducibility with "
                                     "corresponding experiment is not ensured.")
            elif regime not in ["auto-save", "locked"]:
                raise ValueError("'overwriting_regime' is a special parameter. "
                                 "It can only be set to 'auto-save'(default), "
                                 "'locked' or 'unsafe'.")
            self.config_metadata["overwriting_regime"] = regime

            self._former_saving_time = float(value.split("(")[-1].split(")")[0])
            self.set_pre_processing(False)
            return

        # ...do not accept other protected attributes to be merged...
        if key in self._protected_attributes:
            raise RuntimeError(f"Error : '{key}' is a protected name and cannot "
                               "be used as a parameter name.")

        # ... otherwise, process the data normally :

        # If we are merging a parameter into a previously defined config...
        if not any(state.startswith("setup") for state in self._state):
            self._merge_item(key, value)

        # ... or if we are creating a config for the first time and
        # are adding non-existing parameters to it
        else:
            self._add_item(key, value)

    def _merge_item(self, key: str, value: Any) -> None:
        """ Method called by _process_item_to_merge_or_add if the value should be merged and not added. This method
        ultimately performs all merges in the config. """
        if "*" in key:
            to_merge = {}
            for param in self.get_parameter_names(deep=True):
                if compare_string_pattern(param, key):
                    to_merge[param] = value
            if self._verbose:
                if not to_merge:
                    YAECS_LOGGER.warning(f"WARNING : parameter '{key}' will be ignored : it does not match any existing"
                                         " parameter.")
                else:
                    YAECS_LOGGER.info(f"Pattern parameter '{key}' will be merged into the following matched "
                                      f"parameters : {list(to_merge.keys())}.")
            self.init_from_config(to_merge)
        elif "." in key:
            name, new_key = key.split(".", 1)
            try:
                sub_config = getattr(self, "___" + name if name in self._methods else name)
            except AttributeError as exception:
                raise AttributeError(f"ERROR : parameter '{key}' cannot be merged : "
                                     f"it is not in the default '{self.get_name().upper()}' "
                                     f"config.\n{self._did_you_mean(key)}") from exception

            if isinstance(sub_config, _ConfigurationBase):
                sub_config.init_from_config({new_key: value})
            else:
                did_you_mean_message = self._did_you_mean(
                    key.split('.')[0], filter_type=self.__class__, suffix=key.split('.', 1)[1])
                raise TypeError(f"Failed to set parameter '{key}' : '{key.split('.')[0]}'"
                                f" is not a sub-config.\n{did_you_mean_message}")
        else:
            try:
                old_value = getattr(self, "___" + key if key in self._methods else key)
            except AttributeError as exception:
                raise AttributeError(f"ERROR : parameter '{key}' cannot be merged : "
                                     f"it is not in the default '{self.get_name().upper()}' "
                                     f"config.\n{self._did_you_mean(key)}") from exception
            if isinstance(old_value, _ConfigurationBase):
                if isinstance(value, _ConfigurationBase):
                    self.unset_sub_config(value)
                    old_value.init_from_config(value.get_dict(deep=False, pre_post_processing_values=False))
                elif isinstance(value, dict):
                    old_value.init_from_config(value)
                else:
                    raise TypeError(f"Trying to set sub-config '{old_value.get_name()}'\n"
                                    f"with non-config element '{value}'.\n"
                                    "This replacement cannot be performed.")
            else:
                if isinstance(value, _ConfigurationBase):
                    self.unset_sub_config(value)
                    for sub_config in value.get_all_linked_sub_configs():
                        self.unset_sub_config(sub_config)
                    value = value.get_dict(deep=True, pre_post_processing_values=False)
                if self._verbose:
                    YAECS_LOGGER.debug(f"Setting '{key}' : \nold : '{old_value}' \n"
                                       f"new : '{value}'.")
                object.__setattr__(self, "___" + key if key in self._methods else key,
                                   self._process_parameter(key, value, "pre"))
                if key not in self._modified_buffer:
                    self._modified_buffer.append(key)

    def _add_item(self, key: str, value: Any) -> None:
        """ Method called by _process_item_to_merge_or_add if the value should be added and not merged. This method
        ultimately performs all additions to the config. """
        if self._state[0].split(";")[0] == "setup" and "*" in key:
            raise ValueError("The '*' character is not authorised in the default "
                             f"config ({key}).")
        if "." in key and "*" not in key.split(".")[0]:
            name = key.split(".")[0]
            param_name = "___" + name if name in self._methods else name
            try:
                sub_config = getattr(self, param_name)
            except AttributeError:
                self._add_sub_config(name, param_name, {key.split(".", 1)[1]: value})
            else:
                if isinstance(sub_config, _ConfigurationBase):
                    sub_config.init_from_config({key.split(".", 1)[1]: value})
                else:
                    did_you_mean = self._did_you_mean(
                        key.split(".")[0], filter_type=self.__class__, suffix=key.split(".", 1)[1],
                    )
                    raise TypeError("Failed to set parameter "
                                    f"'{key}' : '{key.split('.')[0]}' "
                                    f"is not a sub-config.\n{did_you_mean}")
        else:
            param_name = "___" + key if key in self._methods else key
            try:
                if key != "config_metadata":
                    _ = getattr(self, param_name)
                    raise RuntimeError(f"ERROR : parameter '{key}' was set twice.")
            except AttributeError:
                if key in self._methods and self._verbose:
                    YAECS_LOGGER.warning(f"WARNING : '{key}' is the name of a method in the Configuration object.\n"
                                         f"Your parameter was initialised anyways, under the name ___{key}. You can "
                                         f"access it via config.___{key} or config['{key}'].")
                if isinstance(value, _ConfigurationBase):
                    self._add_sub_config(key, param_name, {k: value[k] for k in value.get_parameter_names(False)},
                                         value.get_type_hints())
                else:
                    if (self._state[0].split(";")[0] == "setup"
                            and [i.split(";")[0] for i in self._state].count("setup") < 2):
                        preprocessed_parameter = self._process_parameter(key, value, "pre")
                    else:
                        preprocessed_parameter = value
                    object.__setattr__(self, param_name, preprocessed_parameter,)
                    if key not in self._modified_buffer:
                        self._modified_buffer.append(key)

    def _add_sub_config(self, name: str, name_in_config: str, content: dict,
                        type_hints: Optional[Dict[str, TypeHint]] = None):
        """ Method called by _add_item to add a sub-config to a config. First an empty config is created, then its
        values are added with its init_from_config method. """
        # This has to be performed in two steps, otherwise
        # the param inside the new sub-config does not get
        # pre-processed.
        object.__setattr__(
            self, name_in_config,
            self._get_instance(
                name=name,
                overwriting_regime=(self._main_config.config_metadata["overwriting_regime"]),
                config_path_or_dictionary={}, state=self._state,
                nesting_hierarchy=self._nesting_hierarchy + [name_in_config],
                main_config=self._main_config, verbose=self._verbose
            ),
        )
        # Now, outside the nested "setup" state during __init__,
        # pre-processing is active
        type_hints_to_transfer = []
        prefix = self._get_full_path(name) + "."
        for type_hint in self._main_config.get_type_hints():
            if type_hint.startswith(prefix):
                type_hints_to_transfer.append(type_hint)
        for type_hint in type_hints_to_transfer:
            self[name].add_type_hint(type_hint[len(prefix):], self._main_config.get_type_hint(type_hint))
            self._main_config.remove_type_hint(type_hint)
        if type_hints is not None:
            for key, value in type_hints.items():
                self[name].add_type_hint(key, value)
        self[name].init_from_config(content)
        self[name].config_metadata["config_hierarchy"] += [content]
        self.set_sub_config(self[name])

    def _gather_command_line_dict(self, to_merge: Optional[Union[List[str], str]] = None) -> Dict[str, Any]:
        """ Method called automatically at the end of each constructor to gather all parameters from the command line
        into a dictionary. This dictionary is then merged. """

        if to_merge is not None:
            if isinstance(to_merge, list):
                to_merge = " ".join(to_merge)
            list_to_merge = get_quasi_bash_sys_argv(to_merge)
        else:
            list_to_merge = sys.argv

        # Setting the config to operational mode in case this
        # is called manually
        object.__setattr__(self, "_operating_creation_or_merging", True)

        # Gather parameters, their values and their types
        to_merge = {}  # {param_name: [former_value, new_value, type_forcing], ...}
        found_config_path = not bool(self._from_argv)
        in_param = []
        un_matched_params = []
        for element in list_to_merge:
            if element.startswith("--") and (found_config_path or element != self._from_argv):
                if "=" in element:
                    pattern, value = element[2:].split("=", 1)
                    value = value if value != "" else None
                else:
                    pattern, value = element[2:], None
                in_param = []
                for parameter in self.get_parameter_names(deep=True):
                    if compare_string_pattern(parameter, pattern):
                        in_param.append(parameter)
                        to_merge[parameter] = value
                if not in_param:
                    un_matched_params.append(pattern)
            elif element.startswith("--"):
                in_param = []
                found_config_path = True
            elif in_param and to_merge[in_param[0]] is None:
                for parameter in in_param:
                    to_merge[parameter] = element
            elif in_param:
                for parameter in in_param:
                    to_merge[parameter] = f"{to_merge[parameter]} {element}"

        if un_matched_params and self._verbose:
            YAECS_LOGGER.warning(f"WARNING : parameters {un_matched_params}, encountered while merging params from the "
                                 f"command line, do not match any param in the config. They will not be merged.")

        # Infer types, then return
        return {key: yaml.safe_load("true" if val is None else val) for key, val in to_merge.items()}

    def _post_process_modified_parameters(self) -> None:
        """ This method is called at the end of a config creation or merging operation. It applies post-processing to
        all parameters modified by this operation. If a parameter is converted into a non-native YAML type, also keeps
        its former value in memory for saving purposes. """
        modified = [
            self._get_full_path(self._modified_buffer.pop(0))
            for _ in range(len(self._modified_buffer))
        ]
        for subconfig in self.get_all_linked_sub_configs():
            modified_buffer = subconfig.get_modified_buffer()
            for _ in range(len(modified_buffer)):
                modified.append(".".join(subconfig.get_nesting_hierarchy() + [modified_buffer.pop(0)]))
        processors = [(proc if isinstance(proc, Callable)
                       else self._assigned_as_yaml_tags[proc[len("_tagged_method_"):]][0])
                      for proc in self._post_processing_functions.values()]
        orders = sorted(list({get_order(func) for func in processors}))
        splits = [name.split(".")[len(self._nesting_hierarchy):] for name in modified]
        names = [(".".join(s), ".".join(s[:-1] + ["___" + s[-1]]) if s[-1] in self._methods else ".".join(s))
                 for s in splits]
        for order in orders:
            for name, set_name in names:
                recursive_set_attribute(self, set_name, self._process_parameter(name, self[name], "post", order))
        post_processed = [param for param in modified if param in self._pre_postprocessing_values]
        if post_processed and self._verbose:
            YAECS_LOGGER.info(f"Performed post-processing for modified parameters {post_processed}.")

    def _prepare_processing_functions(self, processing_type: str) -> None:
        """ Sets self._pre/post_processing_functions from the user-provided functions. """
        processing_functions = {**getattr(self, f"parameters_{processing_type}_processing")(),
                                **getattr(self, f"_added_{processing_type}_processing")()}
        for key, value in processing_functions.items():

            if not isinstance(value, (Callable, Iterable)):
                raise TypeError(f"Invalid {processing_type}-processing functions defined for param '{key}' : "
                                "the function should be declared as either a function or an iterable of functions, "
                                "optionally containing one order value.")

            if isinstance(value, Iterable) and not (isinstance(value, str) and value.startswith("_tagged_method_")):
                if any(not isinstance(element, (Callable, Real)) for element in value):
                    raise TypeError(f"Invalid {processing_type}-processing functions defined for param '{key}' : "
                                    "if function is declared as iterable, only functions and one order value can "
                                    "be provided.")
                order = [i for i in value if isinstance(i, Real)]
                if len(order) > 1:
                    raise ValueError(f"Ambiguous order for {processing_type}-processing functions defined for param "
                                     f"'{key}' : multiple orders defined ({order}).")
                processing_function = compose(*[i for i in value if isinstance(i, Callable)])
                order = order[0] if order else get_order(processing_function)
                set_function_attribute(processing_function, "order", order)

            else:
                processing_function = value

            self.add_processing_function(key, processing_function, processing_type)

    @update_state("processing;_name")
    def _process_parameter(self, name: str, parameter: Any, processing_type: str, order: Optional[Real] = None) -> Any:
        """ This method checks if a processing function has been defined for given name, then returns the processed
        value when that is the case. """
        if processing_type not in ["pre", "post"]:
            raise ValueError(f"Unknown processing_type : '{processing_type}'. Valid types are 'pre' or 'post'.")
        total_name = self._get_full_path(name)
        main = self.get_main_config()
        processors = [proc for key, proc in getattr(self, f"_{processing_type}_processing_functions").items()
                      if compare_string_pattern(total_name, key)]
        processors = [(proc if isinstance(proc, Callable)
                       else self._assigned_as_yaml_tags[proc[len("_tagged_method_"):]][0]) for proc in processors]
        processors = sorted([p for p in processors if order is None or get_order(p) == order], key=get_order)
        if processing_type == "pre":
            main.remove_value_before_postprocessing(total_name)
        if main.get_master_switch(processing_type):
            old_value = None
            if processing_type == "pre":
                self.check_type(main.get_type_hint(total_name))(parameter)
            else:
                old_value = copy.deepcopy(parameter)
            was_processed = bool(processors)
            for processor in processors:
                try:
                    parameter = processor(parameter)
                except Exception:
                    YAECS_LOGGER.error(f"ERROR while {processing_type}-processing param '{total_name}' :")
                    raise
            if processing_type == "pre" and not is_type_valid(parameter, _ConfigurationBase):
                raise RuntimeError(f"ERROR while pre-processing param '{total_name}' : "
                                   "pre-processing functions that change the type of a "
                                   "param to a non-native YAML type are forbidden because "
                                   "they cannot be saved. Please use a parameter "
                                   "post-processing instead.")
            if processing_type == "post" and was_processed:
                main.save_value_before_postprocessing(self._get_full_path(name), old_value)
        elif processing_type == "pre":
            for processor in processors:
                if processor.__name__.startswith("yaecs_config_hook__"):
                    for hook_name in processor.__name__.split("__")[1].split(","):
                        self.add_currently_processed_param_as_hook(hook_name)
        return parameter

    @staticmethod
    def _superficial_dict_cleanup(dictionary: dict) -> dict:
        """ After a YAML loading operation, cleans up composed keys of Configuration items. """
        return {(k.split(".")[0] if isinstance(v, _ConfigurationBase) else k): v for k, v in dictionary.items()}



"""
Reactive Reality Machine Learning Config System - Configuration object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import sys
import time
from typing import Callable, Dict, List, Optional

from ..yaecs_utils import ConfigInput, ConfigDeclarator, format_str, get_config_from_argv
from .config_base import _ConfigurationBase

YAECS_LOGGER = logging.getLogger(__name__)


class Configuration(_ConfigurationBase):
    """
    Superclass for YAECS configurations. The superclass implements constructors, while the behaviour is spread
    across the parent classes in the following way :

    * _ConfigurationBase implements the most basic functionalities such as creation and merging operations. It inherits
      from all other parent classes as Mixins ;
    * ConfigHooksMixin implements processing functions whose name start with `register_as_` and are decorated by the
      yaecs_utils.hook decorator. Users can use those processing either as pre- or post-processing functions, which will
      have the added effect of tagging the processed parameters as playing a certain pre-defined role in the config ;
    * ConfigGettersMixin implements public getters to access some private attributes as well as other values which
      require boilerplate code to be rigorously queried, such as user-defined parameters ;
    * ConfigSettersMixin implements a few setters which the config uses internally to manipulate private attributes of
      other Configuration instances (such as sub-configs) ;
    * ConfigConvenienceMixin implements all functions which are not of central importance to the behaviour of the config
      but wrap convenient functionalities which can be used either internally of by the user.

    """

    def __init__(self, name: str = "main", overwriting_regime: str = "auto-save",
                 config_path_or_dictionary: Optional[ConfigDeclarator] = None,
                 nesting_hierarchy: Optional[List[str]] = None, state: Optional[List[str]] = None,
                 main_config: Optional['Configuration'] = None, variation: Optional[str] = None,
                 do_not_pre_process: bool = False, do_not_post_process: bool = False, verbose: bool = True,
                 **kwargs):
        """
        Should never be called directly by the user. Please use one of the constructors instead (load_config,
        build_from_configs, build_from_argv), or the utils.make_config convenience function.

        :param name: name for the config or sub-config
        :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
            is not guaranteed).
        :param config_path_or_dictionary: path or dictionary to create the config from
        :param nesting_hierarchy: list containing the names of all the configs in the sub-config chain leading to this
            config
        :param state: processing state used for state tracking and debugging
        :param main_config: main config corresponding to this sub-config, or None if this config is the main config
        :param variation: the name of the variation being created
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        :param verbose: if set to false, message logging by this config will be deactivated
        :raises ValueError: if the overwriting regime is not valid
        :return: none
        """

        # Set protected attributes
        object.__setattr__(self, "_operating_creation_or_merging", True)
        self._state = [] if state is None else state
        self._main_config = self if main_config is None else main_config
        self._methods = [name for name in dir(self)
                         if name not in ["_operating_creation_or_merging", "_main_config", "_state"]]
        self._configuration_variations = []
        self._configuration_variations_names = []
        self._grids = []
        self._name = name
        self._nesting_hierarchy = ([] if nesting_hierarchy is None else list(nesting_hierarchy))
        self._variation_name = (variation if main_config is None else main_config.get_variation_name())
        self._verbose = verbose
        kwargs = {"do_not_pre_process": do_not_pre_process, "do_not_post_process": do_not_post_process, **kwargs}
        super().__init__(**kwargs)
        self._protected_attributes = list(self.__dict__) + ["_protected_attributes"]

        # Set config metadata
        self.config_metadata = {
            "saving_time": time.time(),
            "config_hierarchy": [],
            "overwriting_regime": (overwriting_regime if main_config is None
                                   else main_config.config_metadata["overwriting_regime"]),
        }
        if self.config_metadata["overwriting_regime"] not in ["unsafe", "auto-save", "locked"]:
            raise ValueError("'overwriting_regime' needs to be either 'auto-save', "
                             "'locked' or 'unsafe'.")

        # Initialise config
        self._state.append(f"setup;{self._name}")
        config_path_or_dictionary = (self.get_default_config_path()
                                     if config_path_or_dictionary is None else config_path_or_dictionary)
        self.init_from_config(config_path_or_dictionary)

        # Find config hierarchy from parents and set it
        param = None
        if self._nesting_hierarchy:
            for index in range(len(self._nesting_hierarchy) - 1):
                param = main_config.get('.'.join(self._nesting_hierarchy[:index + 1]), main_config)
                if not isinstance(param, _ConfigurationBase):
                    param = None  # encountered in temporary configs when merging a deep dict
                    break
        self.config_metadata["config_hierarchy"] = [] if param is None else param.config_metadata['config_hierarchy']
        self.config_metadata["config_hierarchy"] += [config_path_or_dictionary]

        # Checkup and post-setup operations
        if not self._nesting_hierarchy:
            self._check_for_unlinked_sub_configs()
        if main_config is None:
            self.set_pre_processing(True)
        self._state.pop(-1)
        self._operating_creation_or_merging = False

    @classmethod
    def load_config(cls, *configs: ConfigInput, default_config_path: Optional[ConfigDeclarator] = None,
                    overwriting_regime: str = "auto-save", do_not_merge_command_line: bool = False,
                    do_not_pre_process: bool = False, do_not_post_process: bool = False,
                    **kwargs) -> 'Configuration':
        """
        First creates a config using the default config, then merges config_path into it. If config_path is a list,
        successively merges all configs in the list instead from index 0 to the last.

        :param configs: config's path or dictionary, or list of default config's paths or dictionaries to merge
        :param default_config_path: default config's path or dictionary
        :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
            is not guaranteed).
        :param do_not_merge_command_line: if True, does not try to merge the command line parameters
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        :param kwargs: additional parameters to pass to Configuration
        :return: instance of Configuration object containing the desired config
        """
        # This needs to access protected members of the Configuration class several times to prepare the config
        # pylint: disable=protected-access
        default_config_path = (cls.get_default_config_path() if default_config_path is None else default_config_path)
        if kwargs.get("verbose", True):
            YAECS_LOGGER.info(f"Building config from default : {format_str(default_config_path)}")
        config = cls(config_path_or_dictionary=default_config_path, overwriting_regime=overwriting_regime,
                     do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process, **kwargs)
        if configs and isinstance(configs[0], list):
            configs = configs[0]
        for path in configs:
            config._merge(path, do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process)
        if not do_not_merge_command_line:
            to_merge = config._gather_command_line_dict()
            if to_merge:
                config._merge(to_merge, do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process,
                              source="command line")
        config._post_process_modified_parameters()
        config.set_post_processing(True)
        return config

    @classmethod
    def build_from_configs(cls, *configs: ConfigInput, overwriting_regime: str = "auto-save",
                           do_not_merge_command_line: bool = False, do_not_pre_process: bool = False,
                           do_not_post_process: bool = False, **kwargs) -> 'Configuration':
        """
        First creates a config using the first config provided (or the first config in the provided list), then merges
        the subsequent configs into it from index 1 to the last.

        :param configs: config's path or dictionary, or list of default config's paths or dictionaries to merge
        :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
            is not guaranteed).
        :param do_not_merge_command_line: if True, does not try to merge the command line parameters
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        :param kwargs: additional parameters to pass to Configuration
        :raises TypeError: if configs is invalid
        :return: instance of Configuration object containing the desired config
        """
        if not configs or (isinstance(configs[0], list) and not configs[0]):
            raise TypeError("build_from_configs needs to be called with "
                            "at least one config.")
        if isinstance(configs[0], list) and len(configs) == 1:
            configs = configs[0]
        elif isinstance(configs[0], list):
            raise TypeError(f"Invalid argument : '{configs}'\n"
                            "please use build_from_configs([cfg1, cfg2, ...]) or "
                            "build_from_configs(cfg1, cfg2, ...)")
        return cls.load_config(list(configs[1:]), default_config_path=configs[0], overwriting_regime=overwriting_regime,
                               do_not_merge_command_line=do_not_merge_command_line,
                               do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process, **kwargs)

    @classmethod
    def build_from_argv(cls, *configs: ConfigInput, fallback: Optional[ConfigInput] = None, pattern: str = "--config",
                        default_config_path: Optional[ConfigDeclarator] = None, overwriting_regime: str = "auto-save",
                        do_not_merge_command_line: bool = False, do_not_pre_process: bool = False,
                        do_not_post_process: bool = False, **kwargs) -> 'Configuration':
        """
        Assumes a pattern of the form '--config <path_to_config>' or '--config [<path1>,<path2>,...]' in sys.argv (the
        brackets are optional), and builds a config from the specified paths by merging them into the default config in
        the specified order.

        :param configs: config's path or dictionary, or list of config paths or dictionaries to merge. Those will be
            merged to the default config before the config from the command line.
        :param fallback: config path or dictionary, or list of config paths or dictionaries to fall back to if no config
            was detected in the argv
        :param pattern: pattern to look for in sys.argv
        :param default_config_path: default config's path or dictionary
        :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
            is not guaranteed).
        :param do_not_merge_command_line: if True, does not try to merge the command line parameters
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        :param kwargs: additional parameters to pass to Configuration
        :raises TypeError: if --config is not found and no fallback
        :return: instance of Configuration object containing the desired config
        """
        configs_from_argv: List[ConfigDeclarator] = get_config_from_argv(pattern=pattern, fallback=fallback)

        return cls.load_config(*configs, *configs_from_argv, default_config_path=default_config_path,
                               overwriting_regime=overwriting_regime,
                               do_not_merge_command_line=do_not_merge_command_line,
                               do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process,
                               from_argv=pattern if pattern in sys.argv else "", **kwargs)

    def create_variations(self) -> List['Configuration']:
        """
        Creates a list of configs that are derived from the current config using the internally tracked variations and
        grids registered via the corresponding functions (register_as_config_variations and register_as_grid).

        :raises TypeError: if grid dimension is empty or not registered
        :return: the list of configs corresponding to the tracked variations
        """
        variations_names_to_use = [i[0] for i in self._configuration_variations]
        variations_names_to_use_changing = [i[0] for i in self._configuration_variations]
        variations_to_use = [i[1] for i in self._configuration_variations]
        variations_to_use_changing = [i[1] for i in self._configuration_variations]
        variations = []
        variations_names = []

        def _add_new_names(new_names, dim, var_ind: int):
            for vari_name in self._configuration_variations_names:
                if vari_name[0] == dim:
                    new_names.append(names_to_add[var_ind] + "+" + vari_name[0] + "_" + vari_name[1][index])
            return new_names

        # Adding grids
        for grid in self._grids:
            grid_to_add = []
            names_to_add = []
            for dimension in grid:
                if dimension not in variations_names_to_use:
                    raise TypeError(f"Grid element '{dimension}' is an empty list or "
                                    "not a registered variation configuration.")
                if dimension in variations_names_to_use_changing:
                    index = variations_names_to_use_changing.index(dimension)
                    variations_names_to_use_changing.pop(index)
                    variations_to_use_changing.pop(index)
                if not grid_to_add:
                    grid_to_add = [[i] for i in variations_to_use[variations_names_to_use.index(dimension)]]
                    for var_name in self._configuration_variations_names:
                        if var_name[0] == dimension:
                            names_to_add = [var_name[0] + "_" + i for i in var_name[1]]
                else:
                    new_grid_to_add = []
                    new_names_to_add = []
                    for var_index, current_variation in enumerate(grid_to_add):
                        for index in range(len(variations_to_use[variations_names_to_use.index(dimension)])):
                            new_grid_to_add.append(
                                current_variation
                                + [variations_to_use[variations_names_to_use.index(dimension)][index]])
                            new_names_to_add = _add_new_names(new_names_to_add, dimension, var_index)
                    grid_to_add = [list(var) for var in new_grid_to_add]
                    names_to_add = list(new_names_to_add)
            variations = variations + grid_to_add
            variations_names = variations_names + names_to_add

        # Adding remaining non-grid variations
        for var_idx, remaining_variations in enumerate(variations_to_use_changing):
            for variation_index, variation in enumerate(remaining_variations):
                variations.append([variation])
                name = variations_names_to_use_changing[var_idx]
                for var_name in self._configuration_variations_names:
                    if var_name[0] == name:
                        variations_names.append(name + "_" + var_name[1][variation_index])

        # Creating configs
        variation_configs = []
        for variation_index, variation in enumerate(variations):
            variation_configs.append(
                self.__class__.load_config(self.config_metadata["config_hierarchy"][1:] + variation,
                                           default_config_path=self.config_metadata["config_hierarchy"][0],
                                           overwriting_regime=self.config_metadata["overwriting_regime"],
                                           do_not_merge_command_line=True, variation=variations_names[variation_index],
                                           verbose=False))
            variation_configs[-1].set_variation_name(variations_names[variation_index], deep=True)
        if not variation_configs:
            variation_configs = [self]
        elif self._verbose:
            YAECS_LOGGER.info(f"Created {len(variation_configs)} variation{'s' if len(variation_configs) > 1 else ''} "
                              f"from registered variation parameters.")
        return variation_configs

    @staticmethod
    def get_default_config_path() -> str:
        """
        Returns the path to the default config of the project. This function must be implemented at project-level.

        :return: string corresponding to the path to the default config of the project
        """
        raise NotImplementedError

    def parameters_pre_processing(self) -> Dict[str, Callable]:
        """
        Returns a dictionary where the keys are parameter names (supporting the '*' character as a replacement for any
        number of characters) and the items are functions. The pre-processing functions need to take a single argument
        and return the new value of the parameter after pre-processing. During pre-processing, all parameters
        corresponding to the parameter name are passed to the corresponding function and their value is replaced by the
        value returned by the corresponding function. This function must be implemented at project-level.

        Using this is advised when an action needs to  happen during the ongoing creation or merging operation, such as
        the register_as_additional_config_file processing function, or when a parameter is stored on disk using a format
        that you would prefer to not be used within the config, as the pre-processing function will be performed before
        the parameter even enters the Configuration object.

        Conversions to non-YAML-readable types are forbidden using pre-processing. Please use post-processing for those
        functions.

        :return: dictionary of the pre-processing functions
        """
        raise NotImplementedError

    def parameters_post_processing(self) -> Dict[str, Callable]:
        """
        Returns a dictionary where the keys are parameter names (supporting the '*' character as a replacement for any
        number of characters) and the values are functions. The post-processing functions need to take a single argument
        and return the new value of the parameter after post-processing. After any creation or merging operation,
        parameters which were modified by said operation get post-processed according to the specified functions. This
        function can be implemented at project-level.

        Using this is advised for type-changing processing functions or processing functions which have consequences
        beyond the value of that parameter (for example if they rely on another parameter being initialised, or if they
        would create directories). A notable exception to this rule of thumb is the register_as_additional_config_file
        processing function, which should almost always be called as a pre-processing function. Since the value of the
        parameter should no longer change after post-processing, you can also use post-processing to check if the value
        of the parameter has the correct type or is in the correct range.

        :return: dictionary of the post-processing functions
        """
        raise NotImplementedError

    def set_variation_name(self, name: str, deep: bool = False) -> None:
        """
        Sets the variation index of the config. This function is not intended to be used by the user.

        :param name: index to set the variation index with
        :param deep: whether to also recursively set the variation name of all sub-configs
        """
        object.__setattr__(self, "_variation_name", name)
        if deep:
            for subconfig in self.get_all_sub_configs():
                if ".".join(subconfig.get_nesting_hierarchy()).startswith(".".join(self.get_nesting_hierarchy())):
                    object.__setattr__(subconfig, "_variation_name", name)

    @classmethod
    def _get_instance(cls, name: str = "main", overwriting_regime: str = "auto-save",
                      config_path_or_dictionary: Optional[ConfigDeclarator] = None,
                      nesting_hierarchy: Optional[List[str]] = None, state: Optional[List[str]] = None,
                      main_config: Optional['Configuration'] = None, variation: Optional[str] = None,
                      do_not_pre_process: bool = False, do_not_post_process: bool = False, verbose: bool = True,
                      **kwargs) -> 'Configuration':
        """ Used by parent classes to spawn new instances of the superclass. """
        return cls(name=name, overwriting_regime=overwriting_regime,
                   config_path_or_dictionary=config_path_or_dictionary, nesting_hierarchy=nesting_hierarchy,
                   state=state, main_config=main_config, variation=variation, do_not_pre_process=do_not_pre_process,
                   do_not_post_process=do_not_post_process, verbose=verbose, **kwargs)



"""
Reactive Reality Machine Learning Config System - Configuration object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import importlib.util
from typing import Any, Union

from ..user_utils import tqdm_file

try:
    _NEW = True
    import pytorch_lightning.callbacks.progress.tqdm_progress.TQDMProgressBar as ProgressBar
except ImportError:
    _NEW = False
    from pytorch_lightning.callbacks.progress import ProgressBar
if importlib.util.find_spec("ipywidgets") is not None:
    from tqdm.auto import tqdm as _tqdm  # pylint: disable=import-error
else:
    from tqdm import tqdm as _tqdm  # pylint: disable=import-error


_PAD_SIZE = 5


class Tqdm(_tqdm):
    """ Re-implements pytorch-lightning's TQDM loading bars to make them more YAECS-friendly. """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # pylint: disable=useless-super-delegation
        """Custom tqdm progressbar where we append 0 to floating points/strings to prevent the progress bar from
        flickering."""
        # this just to make the make docs happy, otherwise it pulls docs which has some issues...
        super().__init__(*args, **kwargs)

    @staticmethod
    def format_num(n: Union[int, float, str]) -> str:
        """Add additional padding to the formatted numbers."""
        should_be_padded = isinstance(n, (float, str))
        if not isinstance(n, str):
            n = _tqdm.format_num(n)
            assert isinstance(n, str)
        if should_be_padded and "e" not in n:
            if "." not in n and len(n) < _PAD_SIZE:
                try:
                    _ = float(n)
                except ValueError:
                    return n
                n += "."
            n += "0" * (_PAD_SIZE - len(n))
        return n


class TrackerFriendlyBar(ProgressBar):
    """
    Progress bar class to use instead of TQDMProgressBar (formerly ProgressBar) in pytorch-lightning when yaecs is used
    as a tracker.
    Usage : https://pytorch-lightning.readthedocs.io/en/stable/common/progress_bar.html
    """
    def init_sanity_tqdm(self) -> Tqdm:
        """Override this to customise the tqdm bar for the validation sanity run."""
        tqdm_bar = Tqdm(
            desc=self.sanity_check_description if _NEW else "Validation sanity check",
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=False,
            dynamic_ncols=True,
            file=tqdm_file(),
        )
        return tqdm_bar

    def init_train_tqdm(self) -> Tqdm:
        """Override this to customise the tqdm bar for training."""
        tqdm_bar = Tqdm(
            desc=self.train_description if _NEW else "Training",
            initial=self.train_batch_idx,
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=True,
            file=tqdm_file(),
            smoothing=0,
        )
        return tqdm_bar

    def init_predict_tqdm(self) -> Tqdm:
        """Override this to customise the tqdm bar for predicting."""
        tqdm_bar = Tqdm(
            desc=self.predict_description if _NEW else "Predicting",
            initial=self.train_batch_idx,
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=True,
            file=tqdm_file(),
            smoothing=0,
        )
        return tqdm_bar

    def init_validation_tqdm(self) -> Tqdm:
        """Override this to customise the tqdm bar for validation."""
        # The main progress bar doesn't exist in `trainer.validate()`
        has_main_bar = self.trainer.state.fn != "validate" if _NEW else self.main_progress_bar is not None
        tqdm_bar = Tqdm(
            desc=self.validation_description if _NEW else "Validating",
            position=(2 * self.process_position + has_main_bar),
            disable=self.is_disabled,
            leave=(not has_main_bar) and _NEW,
            dynamic_ncols=True,
            file=tqdm_file(),
        )
        return tqdm_bar

    def init_test_tqdm(self) -> Tqdm:
        """Override this to customise the tqdm bar for testing."""
        tqdm_bar = Tqdm(
            desc="Testing",
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=True,
            file=tqdm_file(),
        )
        return tqdm_bar



"""
Reactive Reality Machine Learning Config System - Configuration object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import sys
from typing import Callable, Dict, List, Optional, Type, Union

from .config.config import Configuration
from .yaecs_utils import ConfigInput, ConfigDeclarator, TqdmLogger, get_config_from_argv

YAECS_LOGGER = logging.getLogger(__name__)


def tqdm_file():
    """
    Utility function which returns a file to which users can log their TQDM bars to make them YAECS-friendly.

    :return: TQDM file
    """
    return TqdmLogger(logging.getLogger("yaecs.print_catcher"))


def get_template_class(default_config_path: Optional[ConfigDeclarator] = None,
                       pre_processing_dict: Optional[Dict[str, Callable]] = None,
                       post_processing_dict: Optional[Dict[str, Callable]] = None,
                       experiment_path: str = None, tracker_config: str = None,
                       additional_configs_suffix: Optional[str] = None,
                       additional_configs_prefix: Optional[str] = None,
                       variations_suffix: Optional[str] = None, variations_prefix: Optional[str] = None,
                       grids_suffix: Optional[str] = None, grids_prefix: Optional[str] = None) -> Type[Configuration]:
    """
    Creates a template Configuration subclass to use in a small project where little customisation is needed.

    :param default_config_path: path to the default config to use for the template
    :param pre_processing_dict: pre-processing dict to use for the template. If this gets large, consider implementing
        the subclass yourself as this will be clearer and more flexible.
    :param post_processing_dict: post-processing dict to use for the template. If this gets large, consider implementing
        the subclass yourself as this will be clearer and more flexible.
    :param experiment_path: automatically adds relevant pre-processing rules to consider given parameter name the
        experiment path
    :param tracker_config: automatically adds relevant pre-processing rules to consider given parameter name the tracker
        config
    :param additional_configs_suffix: automatically adds relevant pre-processing rules to consider parameter names
        ending with 'additional_configs_suffix' as paths to additional config files
    :param additional_configs_prefix: automatically adds relevant pre-processing rules to consider parameter names
        starting with 'additional_configs_prefix' as paths to additional config files
    :param variations_suffix: automatically adds relevant pre-processing rules to consider parameter names ending with
        'variations_suffix' as config variations
    :param variations_prefix: automatically adds relevant pre-processing rules to consider parameter names starting with
        'variations_prefix' as config variations
    :param grids_suffix: automatically adds relevant pre-processing rules to consider parameter names ending with
        'grids_suffix' as grids
    :param grids_prefix: automatically adds relevant pre-processing rules to consider parameter names starting with
        'grids_prefix' as grids
    :return: a template Configuration subclass
    """

    class Template(Configuration):
        """Template class."""

        @staticmethod
        def get_default_config_path() -> str:
            if default_config_path is not None:
                return default_config_path
            raise ValueError(
                "You are using a template class with no default config path. It can only be used with constructors "
                "which do not require access to a default config. Such constructors include :\n"
                " - build_from_configs\n"
                " - __init__ when config_path_or_dictionary is specified\n"
                " - any of load_config or build_from_argv when a default config is specified.")

        def parameters_pre_processing(self) -> Dict[str, Callable]:
            to_ret = {} if pre_processing_dict is None else pre_processing_dict
            if additional_configs_suffix is not None:
                to_ret[f"*{additional_configs_suffix}"] = self.register_as_additional_config_file
            if additional_configs_prefix is not None:
                to_ret[f"{additional_configs_prefix}*"] = self.register_as_additional_config_file
            if variations_suffix is not None:
                to_ret[f"*{variations_suffix}"] = self.register_as_config_variations
            if variations_prefix is not None:
                to_ret[f"{variations_prefix}*"] = self.register_as_config_variations
            if grids_suffix is not None:
                to_ret[f"*{grids_suffix}"] = self.register_as_grid
            if grids_prefix is not None:
                to_ret[f"{grids_prefix}*"] = self.register_as_grid
            if tracker_config is not None:
                to_ret[f"{tracker_config}"] = self.register_as_tracker_config
            return to_ret

        def parameters_post_processing(self) -> Dict[str, Callable]:
            to_ret = {} if post_processing_dict is None else post_processing_dict
            if experiment_path is not None:
                to_ret[f"{experiment_path}"] = self.register_as_experiment_path
            return to_ret

    return Template


def make_config(*configs: Union[ConfigDeclarator, List[ConfigDeclarator]],
                config_class: Optional[Type[Configuration]] = None,
                pre_processing_dict: Optional[Dict[str, Callable]] = None,
                post_processing_dict: Optional[Dict[str, Callable]] = None,
                experiment_path: str = None, tracker_config: str = None,
                additional_configs_suffix: Optional[str] = None, additional_configs_prefix: Optional[str] = None,
                variations_suffix: Optional[str] = None, variations_prefix: Optional[str] = None,
                grids_suffix: Optional[str] = None, grids_prefix: Optional[str] = None,
                fallback: Optional[ConfigInput] = "{}", pattern: str = "--config",
                **class_building_kwargs) -> Configuration:
    """
    One-liner wrapper to create a config from dicts/strings without the need for declaring a subclass. Useful for
    scripts or jupyter notebooks. Impractical/hacky for larger projects.

    :param configs: dicts and strings defining a config
    :param config_class: class to use to build the configuration. If not provided, use a template instead.
    :param pre_processing_dict: pre-processing dict to use for the template.
        If this gets large, consider implementing the subclass yourself as this will be clearer and more flexible.
        Only used if config_class is not provided.
    :param post_processing_dict: post-processing dict to use for the template.
        If this gets large, consider implementing the subclass yourself as this will be clearer and more flexible.
    :param experiment_path: automatically adds relevant pre-processing rules to consider given parameter name the
        experiment path.
        Only used if config_class is not provided.
    :param tracker_config: automatically adds relevant pre-processing rules to consider given parameter name the tracker
        config.
        Only used if config_class is not provided.
    :param additional_configs_suffix: automatically adds relevant pre-processing rules to consider parameter names
        ending with 'additional_configs_suffix' as paths to additional config files.
        Only used if config_class is not provided.
    :param additional_configs_prefix: automatically adds relevant pre-processing rules to consider parameter names
        starting with 'additional_configs_prefix' as paths to additional config files.
        Only used if config_class is not provided.
    :param variations_suffix: automatically adds relevant pre-processing rules to consider parameter names ending with
        'variations_suffix' as config variations.
        Only used if config_class is not provided.
    :param variations_prefix: automatically adds relevant pre-processing rules to consider parameter names starting with
        'variations_prefix' as config variations.
        Only used if config_class is not provided.
    :param grids_suffix: automatically adds relevant pre-processing rules to consider parameter names ending with
        'grids_suffix' as grids.
        Only used if config_class is not provided.
    :param grids_prefix: automatically adds relevant pre-processing rules to consider parameter names starting with
        'grids_prefix' as grids.
        Only used if config_class is not provided.
    :param fallback: if provided, use this as a fallback when no config is provided in argv. The default value "{}"
        stands for "by default do not merge anything if no merge pattern is found in argv"
    :param pattern: pattern to use to find the config in argv.
    :param class_building_kwargs: same kwargs as those used in all Configuration constructors.
        Only used if config_class is not provided.
    :return: config object
    """

    # Prepare class
    class_args = {
        "pre_processing_dict": pre_processing_dict, "post_processing_dict": post_processing_dict,
        "experiment_path": experiment_path, "tracker_config": tracker_config,
        "additional_configs_suffix": additional_configs_suffix, "additional_configs_prefix": additional_configs_prefix,
        "variations_suffix": variations_suffix, "variations_prefix": variations_prefix,
        "grids_suffix": grids_suffix, "grids_prefix": grids_prefix,
    }
    if config_class is None:
        config_class = get_template_class(**class_args)
    elif any(arg is not None for arg in class_args.values()):
        YAECS_LOGGER.warning("WARNING : The following arguments are not used if config_class is provided :\n"
                             f"{list(class_args.keys())}.")

    # Get configs from argv
    configs_from_argv = get_config_from_argv(pattern=pattern, fallback={} if fallback == "{}" else fallback)
    class_building_kwargs["from_argv"] = pattern if pattern in sys.argv else ""
    if all(c == {} for c in configs_from_argv):
        configs_from_argv = []
        class_building_kwargs["from_argv"] = ""

    return config_class.build_from_configs(*configs, *configs_from_argv, **class_building_kwargs)
